/*
**  WASABI-Hot! version 1.2c   
**
**
**      -- copyright (c) 2001-2004 by Philips Japan, Ltd. -- All rights reserved --
**
**
**      ** This code has been made to check/learn                          ** 
**      **                             the ISP1362/ISP1363 functionalities **
**      ** Release 06-Aug-2004                                             **
**
**      OKANO, Akifumi
**      
**		Application Laboratory, Mobile and Connectivity
**      Semiconductors Div, Philips Japan Ltd.
**      akifumi.okano@philips.com
**      +81-3-3740-4668 
**
*/

/****************************************************************************/
/*	includes																*/
/****************************************************************************/

#include		<string.h>
#include		<mem.h>
#include		<alloc.h>
#include		<dos.h>

#include		"_hc_core/atl_mix.h"
#include		"_hc_core/init.h"
#include		"_hc_core/isr.h"
#include		"_hc_hw/hc_comm.h"
#include		"_hc_hw/hw_acces.h"
#include		"ui.h"


/****************************************************************************/
/*	constants																*/
/****************************************************************************/

#define			TIMEOUT		InitVal_HcATLPTDDoneThresholdTimeOut

#define			AUTO_TRANSFER_RETRY_ENABLE
#define			TEST_VALUE		0


/****************************************************************************/
/*	global vars																*/
/****************************************************************************/

unsigned long			gp_cc_logger[ 16 ]	= { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

transfer_instance		*g_atl_transfer_slot;
transfer_instance		*g_intl_transfer_slot;

unsigned char			g_atl_slot_number_by_block[ 32 ];
buffer_params			g_buffer_parameter[ NUMBER_OF_TRANSFER_TYPES + 1 ];
unsigned char			g_slot_index		= 0;

transfer_instance_ptr	g_processing_block;
unsigned char			g_pp_buffer_sw;
unsigned char			g_pp_pause		= False;
unsigned char			g_pp_block0		= 0xFF;
unsigned char			g_pp_block1		= 0xFF;
unsigned char			g_pp_active		= False;
unsigned long			g_previous_timeout_setting;
unsigned char			g_atl_flag_mask			= False;
unsigned long			g_atl_disabled_time		= 0;


/****************************************************************************/
/*	function prototypes														*/
/****************************************************************************/

void		free_transfer_block( transfer_instance_ptr tip );
void		ATL_IRQ_service( unsigned char dummy );
void		INTL_IRQ_service( unsigned char dummy );
void		atl_scheduler( unsigned char i );
void		intl_scheduler( unsigned char index, unsigned char polling_rate );

void		write_a_PTD_into_1362( transfer_instance_ptr tip );
void		polling_ATL_done_flag( void );
void		handle_transfer_done( transfer_instance_ptr tip );

static void	set_PTD_header( unsigned char *ph, PTD_header_param *php );
static void get_PTD_header( PTD_header_param *php, unsigned char *ph );
void		atlmix_show_all_ptd_header( void );
void		atlmix_show_ptd_header( unsigned char transfer_type, unsigned char index );


/****************************************************************************/
/*	function definitions													*/
/****************************************************************************/

/***************************************
 ***************************************
 **
 **		Paired-PTD
 **
 **       All functions those have name of "atlmix_pp_*" are made to control paired-PTD. 
 **       These are used to optimize the storage access performance
 **
 ***************************************
 ***************************************/

unsigned char atlmix_pp_which_buffer( void )
{
	g_pp_buffer_sw	= g_pp_buffer_sw ? 0 : 1;

	return ( g_pp_buffer_sw );

//	return ( (read_register16( Com16_HcBufferStatus  ) & 0x0400) >> 10 );
}


void atlmix_pp_link_indexes( unsigned char blk0, unsigned char blk1 )
{
	unsigned short	tmp;

	g_pp_pause		= True;
	
	tmp		 = read_register16( Com16_HcBufferStatus  );
	tmp		&= ~0x8;

	write_register16( Com16_HcBufferStatus,  tmp );
	write_register16( Com16_HcBufferStatus,  tmp | 0x0010 );
	write_register16( Com16_HcBufferStatus,  tmp );

	g_pp_buffer_sw	= 1;
	
	g_pp_block0		= blk0;
	g_pp_block1		= blk1;
	
	g_previous_timeout_setting	= read_register16( Com16_HcATLPTDDoneThresholdTimeOut );
//	write_register16( Com16_HcATLPTDDoneThresholdTimeOut, 0xFF );
	write_register16( Com16_HcATLPTDDoneThresholdTimeOut, 0 );
}


void atlmix_pp_start( void )
{
	g_pp_pause		= False;
	g_pp_active		= True;
	write_register16( Com16_HcBufferStatus,  read_register16( Com16_HcBufferStatus  ) | 0x8 );
}


unsigned short atlmix_atl_payload_size( void )
{
	return( read_register16( Com16_HcATLBlkSize ) );
}


void atlmix_pp_unlink_indexes( void )
{
	g_pp_block0		= 0xFF;
	g_pp_block1		= 0xFF;
	g_pp_active		= False;

	write_register16( Com16_HcATLPTDDoneThresholdTimeOut, g_previous_timeout_setting );
}


unsigned char atlmix_pp_atl_active( void )
{
	unsigned char	i;

	if ( !(read_register16( Com16_HcBufferStatus ) & 0x8) )
	{
		if ( (g_atl_disabled_time + 3000L) < gp_sof_counter )
		{
			for ( i = 1; i < g_buffer_parameter[ CTRL_BULK_TRANSFER ].number_of_blocks; i++ )
				if ( g_atl_transfer_slot[ i ].lock == LOCK )
					return ( True );
					
			return ( False );
		}
	}
	
	return ( True );
}


void atlmix_pp_atl_flag_mask( unsigned char flag )
{
	g_atl_flag_mask		= flag;
}


unsigned char atlmix_pp_operation_active( void )
{
	return ( g_pp_active );
}


/***************************************
 ***************************************
 **
 **		Buffer initializations
 **
 ***************************************
 ***************************************/

void atlmix_buffer_setting( unsigned short isochro_single_size, unsigned short interr_payload_size, unsigned short interr_block_size, unsigned short aperio_block_size )
{
	//	
	//	This function will be called at begining of the program.
	//	This function sets parameters for the buffer information.
	//

	buffer_info		bi;
	
	//	Buffer size setting
	
	bi.isochr.single_buffer_size	= isochro_single_size;
	
	bi.interr.payload_size			= interr_payload_size;
	bi.interr.n_of_blocks			= interr_block_size;
	
	bi.aperio.payload_size			= aperio_block_size;


	//	Uninstall interrupt service routines

	isr_install_ATL_IRQ_service_routine( NULL );
	isr_install_INTL_IRQ_service_routine( NULL );


	//	Buffer parameter setting and clear all contents

	hwacces_erase_all();
	atlmix_buffer_parameter_initialize( &bi );


	//	Install interrupt service routines

	isr_install_ATL_IRQ_service_routine( ATL_IRQ_service );
	isr_install_INTL_IRQ_service_routine( INTL_IRQ_service );
}


/***************************************
 ***************************************
 **
 **		Interface for transfers
 **
 ***************************************
 ***************************************/

unsigned char atlmix_get_open_transfer_index( unsigned char transfer_type )
{
	unsigned char	n_of_slots;
	unsigned char	i;
	
	if ( transfer_type == CONTROL_TRANSFER )
	{
		if ( g_atl_transfer_slot[ 0 ].lock != LOCK )
		{
			g_atl_transfer_slot[ 0 ].lock		= LOCK;
			free_transfer_block( &(g_atl_transfer_slot[ 0 ]) );
			return ( 0 );
		}
		else
		{
			return ( NO_OPEN_ATL_TRANSFER_SLOT );
		}
	}
	
	else if ( transfer_type == BULK_TRANSFER )
	{
		n_of_slots	= g_buffer_parameter[ CTRL_BULK_TRANSFER ].number_of_blocks;
	
		for ( i = 1; i < n_of_slots; i++ )
			if ( g_atl_transfer_slot[ i ].lock != LOCK )
			{
				g_atl_transfer_slot[ i ].lock		= LOCK;
				free_transfer_block( &(g_atl_transfer_slot[ i ]) );
				return ( i );
			}
		
		if ( i == n_of_slots )
			return ( NO_OPEN_ATL_TRANSFER_SLOT );
	}
	
	else if ( transfer_type == INTERRUPT_TRANSFER )
	{
		n_of_slots	= g_buffer_parameter[ INTERRUPT_TRANSFER ].number_of_blocks;

		for ( i = 0; i < n_of_slots; i++ )
			if ( g_intl_transfer_slot[ i ].lock != LOCK )
			{
				g_intl_transfer_slot[ i ].lock		= LOCK;
				free_transfer_block( &(g_intl_transfer_slot[ i ]) );
				return ( i );
			}
		
		if ( i == n_of_slots )
			return ( NO_OPEN_ATL_TRANSFER_SLOT );
	}
	
	else
	{
		mprintf( LIGHTRED, ABORT, "atlmix_get_open_transfer_index : Unsupported transfer type   transfer_type = %d   \r\n", transfer_type );
	}
	
	return ( NO_OPEN_ATL_TRANSFER_SLOT );
}


void atlmix_free_transfer_index( unsigned char transfer_type, unsigned char index )
{
	switch ( transfer_type )
	{
		case CONTROL_TRANSFER :
		case BULK_TRANSFER :
		case CTRL_BULK_TRANSFER :
			atlmix_free_transfer( &g_atl_transfer_slot[ index ] );
			break;
		case INTERRUPT_TRANSFER :
			atlmix_free_transfer( &g_intl_transfer_slot[ index ] );
			break;
		default :
			mprintf( LIGHTRED, ABORT, "atlmix_free_transfer_index : Unsupported transfer type   transfer_type = %d   \r\n", transfer_type );
			break;
	}
}


void free_transfer_block( transfer_instance_ptr tip )
{
	tip->target_block_index			= 0xFF;
}


void atlmix_free_transfer( transfer_instance_ptr tip )
{
	unsigned short	skipmap_register;

	disable_INT();	//	##	version1.2c

	skipmap_register	= (tip->transfer_type == INTERRUPT_TRANSFER) ? Com32_HcINTLPTDSkipMap : Com32_HcATLPTDSkipMap;

	write_register32( skipmap_register, read_register32( skipmap_register ) | (0x00000001L << tip->target_block_index) );
		
	if ( tip->epi_ptr )
		(tip->epi_ptr)->transfer_instance		= NULL;

	tip->lock					= UNLOCK;
	tip->data_ptr_base			= NULL;
	tip->data_ptr_start			= NULL;
	tip->data_ptr_end			= NULL;
	tip->epi_ptr				= NULL;
	tip->callback				= NULL;
	tip->transferred_size		= 0;
	tip->direction				= 0;
	tip->completion_code		= 0;
	tip->fatal_error			= 0;
	tip->target_address			= 0;
	tip->target_block_index		= 0xFF;
	tip->retry_count			= 0;
	tip->timeout				= 0xFFFFFFFF;		
	tip->time_of_service		= 0xFFFFFFFF;
	
	free_transfer_block( tip );

	enable_INT();	//	##	version1.2c
}


transfer_instance *atlmix_set_transfer( unsigned char transfer_type, unsigned char polling_rate, unsigned char index, unsigned char *buffer_ptr, unsigned short length, unsigned char direction, endpoint_info_ptr epi_ptr, void (*callback)( endpoint_info_ptr epi_ptr ) )
{
	transfer_instance_ptr	tip;
	unsigned short			intr_en;

	disable_INT();

	switch ( transfer_type )
	{
		case CONTROL_TRANSFER :
		case BULK_TRANSFER :
		
			if ( !(index < g_buffer_parameter[ CTRL_BULK_TRANSFER ].number_of_blocks) )
				return ( NULL );
				
			tip		= &(g_atl_transfer_slot[ index ]);
			
			{
				if ( index == g_pp_block0 )
				{
					tip->paired		= True;
					tip->ping_pong	= 0;
				}
				else if ( index == g_pp_block1 )
				{
					tip->paired		= True;
					tip->ping_pong	= 1;
				}
				else
				{
					tip->paired		= False;
					tip->ping_pong	= 0;
				}
			}
			
			break;
			
		case INTERRUPT_TRANSFER :
		
			if ( !(index < g_buffer_parameter[ INTERRUPT_TRANSFER ].number_of_blocks) )
				return ( NULL );
				
			tip		= &(g_intl_transfer_slot[ index ]);
			break;
			
		default :
			mprintf( LIGHTRED, ABORT, "atlmix_set_transfer : Unsupported transfer type   transfer_type = %d   \r\n", transfer_type );
			break;
	}

	epi_ptr->transfer_instance		= tip;

	tip->data_ptr_base				= buffer_ptr;
	tip->data_ptr_start				= buffer_ptr;
	tip->data_ptr_end				= buffer_ptr + length;
	tip->epi_ptr					= epi_ptr;
	tip->callback					= callback;
	tip->transferred_size			= 0;
	tip->direction					= direction;
	tip->completion_code			= 0;
	tip->fatal_error				= 0;
	tip->transfer_type				= transfer_type;
	tip->retry_count				= 0;
	tip->timeout					= TEST_VALUE;
	
	switch ( transfer_type )
	{
		case CONTROL_TRANSFER :
		case BULK_TRANSFER :

			atl_scheduler( index );
			break;
			
		case INTERRUPT_TRANSFER :

			intl_scheduler( index, polling_rate );
			break;
			
		default :
			mprintf( LIGHTRED, ABORT, "atlmix_set_transfer : Unsupported transfer type   transfer_type = %d   \r\n", transfer_type );
			break;
	}

	enable_INT();

	return ( tip );
}


void re_enable_atl_transfer( transfer_instance_ptr tip )
{
	unsigned short			intr_en;

	disable_INT();

	atl_scheduler( tip->index );

	enable_INT();
}


void atlmix_re_enable_interrupt_transfer( transfer_instance_ptr tip )
{
	unsigned short	v;

	tip->data_ptr_start		= tip->data_ptr_base;
	
	read_direct_access( tip->target_address, (unsigned char *)(&v), 2 );

	v	&= 0x0400;		//	Do not clear "Toggle" information
	v	|= 0x0800;		//	Set "Active" bit to activate this PTD

	write_direct_access( tip->target_address, (unsigned char *)(&v), 2 );		//	Needs to re-write these 2 bytes

	write_register32( Com32_HcINTLPTDSkipMap, read_register32( Com32_HcINTLPTDSkipMap ) & ~(0x00000001L << (tip->index)) );
}



/***************************************
 ***************************************
 **
 **		Schedulers
 **
 ***************************************
 ***************************************/


void atl_scheduler( unsigned char i )
{
	transfer_instance_ptr	tip;

#ifdef USE_SKIPMAP
	unsigned long			skip_map;
#endif

	tip		= &(g_atl_transfer_slot[ i ]);


#if 0
	if ( tip->target_block_index != 0xFF )
		mprintf( YELLOW, ABORT, "atl_scheduler error (%u) : ( tip->target_block_index != 0xFF )\r\n", i );
	
	if ( tip->data_ptr_start == NULL )
		mprintf( YELLOW, ABORT, "atl_scheduler error (%u|%p) : ( tip->data_ptr_start == NULL )\r\n", i, tip );
#endif


#ifdef USE_SKIPMAP
	skip_map	= read_register32( Com32_HcATLPTDSkipMap );

	if ( ~skip_map & (0x1 << i) )
		mprintf( YELLOW, ABORT, "atl_scheduler error (%u) : ( skip_map & (0x1 << i) )\r\n", i );
#endif

	tip->target_block_index					= i;
	g_atl_slot_number_by_block[ i ]			= i;
	
	tip->target_address						= g_buffer_parameter[ CTRL_BULK_TRANSFER ].base_address + ((g_buffer_parameter[ CTRL_BULK_TRANSFER ].payload_size + PTD_HEADER_SIZE) * i);

	write_a_PTD_into_1362( tip );
}


void intl_scheduler( unsigned char index, unsigned char polling_rate )
{
	transfer_instance_ptr	tip;
	unsigned short			tartget_address_1362_buffer;
	unsigned char			i;

	tip		= &(g_intl_transfer_slot[ index ]);


	tip->target_block_index			= index;

	tartget_address_1362_buffer		= g_buffer_parameter[ INTERRUPT_TRANSFER ].base_address + ((g_buffer_parameter[ INTERRUPT_TRANSFER ].payload_size + PTD_HEADER_SIZE) * index);
	
	tip->target_address				= tartget_address_1362_buffer;
	tip->polling_rate				= (128 < polling_rate) ? 128 : polling_rate;

	write_a_PTD_into_1362( tip );
}


void write_a_PTD_into_1362( transfer_instance_ptr tip )
{
	unsigned short		transfer_size;
	unsigned short		payload_size;
	unsigned char		ptd_header_buffer_ptr[ PTD_HEADER_SIZE ];
	PTD_header_param	ph;
	unsigned short		skipmap_register;
	unsigned short		buffer_enable_mask;
	unsigned char		i;

	payload_size		= g_buffer_parameter[ CTRL_BULK_TRANSFER ].payload_size;
	
	transfer_size		= tip->data_ptr_end - tip->data_ptr_start;		//	remaining data size
	transfer_size		= (payload_size < transfer_size) ? payload_size : transfer_size;

	ph.ActualBytes		= 0;


	ph.CompletionCode	= 0xE;

	ph.CompletionCode	= 0;

	ph.Active			= 1;
	ph.Toggle			= (tip->epi_ptr)->toggle;
	ph.MaxPacketSize	= (tip->epi_ptr)->max_packet_size;
	ph.EndpointNumber	= (tip->epi_ptr)->ep_num;
	ph.Last				= 0;
	ph.Speed			= (tip->epi_ptr)->speed;
	ph.TotalBytes		= transfer_size;
	ph.DirectionPID		= tip->direction;
	ph.Format			= (tip->epi_ptr)->tr_type;
	ph.FunctionAddress	= (tip->epi_ptr)->dev_addr;

	if ( tip->transfer_type == INTERRUPT_TRANSFER )
	{
		for ( i = 7; i > 0; i-- )
			if ( (tip->polling_rate) & (0x01 << i) )
				break;
				
		ph.PollingRate		= i;
		ph.StartingFrame	= tip->index;
		ph.Paired			= False;
		ph.PingPong			= 0;
	
		skipmap_register	= Com32_HcINTLPTDSkipMap;
		buffer_enable_mask	= 0x4;
	}
	else
	{
		ph.PollingRate		= 0;
		ph.StartingFrame	= 0;
		ph.Paired			= tip->paired;
		ph.PingPong			= tip->ping_pong;
		
		skipmap_register	= Com32_HcATLPTDSkipMap;
		buffer_enable_mask	= g_pp_pause ? 0x0 : 0x8;
	}
	

	set_PTD_header( ptd_header_buffer_ptr, &ph );			//	write PTD header
	
	if ( read_register16( Com16_HcATLPTDDoneThresholdTimeOut ) )
		tip->timeout			= gp_sof_counter + read_register16( Com16_HcATLPTDDoneThresholdTimeOut );
	else
		tip->timeout			= 0xFFFFFFFF;

	write_direct_access( tip->target_address, ptd_header_buffer_ptr, PTD_HEADER_SIZE );
	
	if ( tip->direction != IN )
		write_direct_access( tip->target_address + PTD_HEADER_SIZE, tip->data_ptr_start, transfer_size );

	write_register32( skipmap_register, read_register32( skipmap_register ) & ~(0x00000001L << tip->target_block_index) );

	write_register16( Com16_HcBufferStatus,  read_register16( Com16_HcBufferStatus  ) |  buffer_enable_mask );
}


/***************************************
 ***************************************
 **
 **		IRQs
 **
 ***************************************
 ***************************************/


void ATL_IRQ_service( unsigned char dummy_for_func_compatibility )
{
	unsigned long	done_map;
	unsigned long	enable_map;
	unsigned long	check_timeout_map;
	unsigned long	timeout_map				= 0L;
	unsigned char	i;

	dummy_for_func_compatibility	= dummy_for_func_compatibility;	//	This line for erase warning

	done_map			=  read_register32( Com32_HcATLPTDDoneMap );	
	enable_map			= ~read_register32( Com32_HcATLPTDSkipMap );
	check_timeout_map	= ~done_map & enable_map;

	if ( check_timeout_map && !g_pp_active )
	{
		for ( i = 0; i < g_buffer_parameter[ CTRL_BULK_TRANSFER ].number_of_blocks; i++ )
			if ( check_timeout_map & (0x1 << i) )
				if ( g_atl_transfer_slot[ g_atl_slot_number_by_block[ i ] ].timeout < gp_sof_counter )
					timeout_map		|= (0x1L << i);

		done_map	|= timeout_map;
	}

	if ( done_map )
	{
		write_register32( Com32_HcATLPTDSkipMap, ~enable_map | done_map );

		if ( 0x00000000 == (enable_map & ~done_map) )
		{
			write_register16( Com16_HcBufferStatus,  read_register16( Com16_HcBufferStatus  ) & ~0x8 );
			
			if ( !g_atl_flag_mask )
				g_atl_disabled_time		= gp_sof_counter;
		}
			
		for ( i = 0; i < g_buffer_parameter[ CTRL_BULK_TRANSFER ].number_of_blocks; i++ )
		{
			if ( done_map & (0x1 << i) )
			{
				if ( (timeout_map & (0x1 << i)) && (read_register16( Com16_HcBufferStatus  ) & 0x8) )
				{
					//	Wait the HC leave from ABORTED ptd-block.
					//	
					//	If it is not aborted ptd-block, software does not need to 
					//	check this. Because HC will not touch this ptd-block 
					//	with no-active flag in pth-header.
					
					//	Version 1.2b : 
					//    When the "HcBufferStatus.ATL_Active (bit position 0x8)" is cleared, 
					//    the HC stops scnanning. In this case, no need to wait the scan goes 
					//    to other slot.
					//    If following loop enabled with ATL_Active==0, the program ca not exit. 
				
					while ( read_register32( Com16_HcATLCurrentActiePTD ) == i )
						;
				}

	//			handle_transfer_done( &(g_atl_transfer_slot[ g_atl_slot_number_by_block[ i ] ]) );
				handle_transfer_done( &(g_atl_transfer_slot[ i ]) );
			}
		}
	}
}


void INTL_IRQ_service( unsigned char dummy_for_func_compatibility )
{
	unsigned long	done_map;
	unsigned char	i;

	dummy_for_func_compatibility	= dummy_for_func_compatibility;	//	This line for erase warning
	
	done_map		= read_register32( Com32_HcINTLPTDDoneMap );
	
	write_register32( Com32_HcINTLPTDSkipMap, read_register32( Com32_HcINTLPTDSkipMap ) | done_map );

	for ( i = 0; i < g_buffer_parameter[ INTERRUPT_TRANSFER ].number_of_blocks; i++ )
		if ( done_map& (0x1 << i) )
			handle_transfer_done( &(g_intl_transfer_slot[ i ]) );
}


/***************************************
 ***************************************
 **
 **		Buffer maintenance for transfer done and timeout
 **
 ***************************************
 ***************************************/

void handle_transfer_done( transfer_instance_ptr tip )
{
	unsigned char		ptd_header_buffer_ptr[ PTD_HEADER_SIZE ];
	PTD_header_param	ph;
	unsigned short		transferred_bytes;

	if ( tip->data_ptr_start == NULL )
		return;

	g_processing_block	= tip;
	
	read_direct_access( tip->target_address, ptd_header_buffer_ptr, PTD_HEADER_SIZE );

	get_PTD_header( &ph, ptd_header_buffer_ptr );

	tip->fatal_error		<<= 1;

	(tip->epi_ptr)->toggle	 = ph.Toggle;

	transferred_bytes		 = ph.ActualBytes;

	if ( tip->direction == IN )
		read_direct_access( tip->target_address + PTD_HEADER_SIZE, tip->data_ptr_start, transferred_bytes );

	tip->data_ptr_start		+= transferred_bytes;
	tip->transferred_size	+= transferred_bytes;
	tip->completion_code	 = ph.CompletionCode;
	tip->retry_count		 = 0;

	gp_cc_logger[ ph.CompletionCode ]++;

	switch ( ph.CompletionCode )
	{
		case 0x0 :		//	No error
			break;
			
		case 0x4 :		//	Stall					- Fatal error
		case 0x9 :		//	DataUnderrun			- Non-fatal error
			break;
			
		case 0x3 :		//	DataToggleMismatch		- Fatal error
		
#ifdef AUTO_TRANSFER_RETRY_ENABLE

			{
				if ( tip->transfer_type == INTERRUPT_TRANSFER )
					atlmix_re_enable_interrupt_transfer( tip );
				else
					re_enable_atl_transfer( tip );
				
				return;
			}
			
#else
			break;
#endif

		case 0x5 :		//	DeviceNotResponding		- Fatal error
		case 0x6 :		//	PIDCheckFailure			- Fatal error
		
			tip->fatal_error	|= 0x1;


#ifdef AUTO_TRANSFER_RETRY_ENABLE


			if ( (tip->fatal_error) != 0x7 )
			{
				if ( tip->transfer_type == INTERRUPT_TRANSFER )
					atlmix_re_enable_interrupt_transfer( tip );
				else
					re_enable_atl_transfer( tip );
				
				return;
			}
#else
			break;
#endif


		case 0x1 :		//	CRC						- Fatal error
		case 0x2 :		//	BitStuffing				- Fatal error
		case 0x7 :		//	UnexpectedPID			- Fatal error
		case 0x8 :		//	DataOverrun				- Non-fatal error
		case 0xC :		//	BufferOverrun			- Non-fatal error
		case 0xD :		//	BufferUnderrun			- Non-fatal error
			break;
			
		default :		//	Other errors
			//	un-documented error, never be happened.

			break;
			
	}


	free_transfer_block( tip );
	tip->data_ptr_start		= NULL;

//	if ( tip->epi_ptr != NULL )

	(*(tip->callback))( tip->epi_ptr );
}


/***************************************
 ***************************************
 **
 **		Utility routines
 **
 ***************************************
 ***************************************/

static void set_PTD_header( unsigned char *ph, PTD_header_param *php )
{
	*(ph + 0)	= php->ActualBytes & 0x0FF;
	
	*(ph + 1)	=     (php->CompletionCode    << 4)
					| (php->Active            << 3) 		
					| (php->Toggle            << 2) 
					| (php->ActualBytes       >> 8);
	
//	(*(ph + 1))	= (*(ph + 1)) | (0xE << 4);
	
	*(ph + 2)	= php->MaxPacketSize & 0xFF;
	
	*(ph + 3)	= 	  (php->EndpointNumber    << 4)
					| (0                      << 3) 		
					| (php->Speed             << 2) 
					| (php->MaxPacketSize     >> 8);
					
	*(ph + 4)	= php->TotalBytes & 0xFF;

//	*(ph + 5)	= 	  (php->DirectionPID      << 2)
//					| (php->TotalBytes        >> 8);
	
	*(ph + 5)	= 	  (php->Paired            << 7)
					| (php->PingPong          << 6)
					| (php->DirectionPID      << 2)
					| (php->TotalBytes        >> 8);
	
	*(ph + 6)	= 	  (php->Format            << 7)
					| (php->FunctionAddress & 0x7F);
	
	*(ph + 7)   =     (php->PollingRate       << 5)
					| (php->StartingFrame & 0x1F);
}


static void get_PTD_header( PTD_header_param *php, unsigned char *ph )
{
	php->ActualBytes		=  *(ph + 0) | ((*(ph + 1) & 0x3) << 8);
	php->CompletionCode		=  *(ph + 1) >> 4;
	php->Active				= (*(ph + 1) >> 3) & 0x01;
	php->Toggle				= (*(ph + 1) >> 2) & 0x01;
	php->MaxPacketSize		=  *(ph + 2) | ((*(ph + 3) & 0x3) << 8);
	php->EndpointNumber		=  *(ph + 3) >> 4;
	php->Last				= (*(ph + 3) >> 3) & 0x01;
	php->Speed				= (*(ph + 3) >> 2) & 0x01;
	php->TotalBytes			=  *(ph + 4) | ((*(ph + 5) & 0x3) << 8);
	php->DirectionPID		= (*(ph + 5) >> 2) & 0x03;
	php->Format				=  *(ph + 6) >> 7;
	php->FunctionAddress	=  *(ph + 6) & 0x7F;
}


unsigned char atlmix_is_fatal_error( transfer_instance *tr_inst_ptr )
{
	unsigned char	completion_code;

	completion_code		= tr_inst_ptr->completion_code;

	return ( (completion_code & 0x8) ? 0 : completion_code );
}

#define DEBUG_CODE
#ifdef DEBUG_CODE
#if 0

void atlmix_show_ptd_header( unsigned char transfer_type, unsigned char index )
{
	unsigned short	p[4];
	unsigned short	target_address;
	
	target_address	= g_buffer_parameter[ transfer_type ].base_address + (g_buffer_parameter[ transfer_type ].payload_size + PTD_HEADER_SIZE) * index;
	
	read_direct_access( target_address, (unsigned char *)p, 8 );
	mprintf( WHITE, CONTINUE, "%d-%d : PTD-h 0x %04X %04X %04X %04X \r\n", transfer_type, index, *(p + 0), *(p + 1), *(p + 2), *(p + 3) );	
}

#else

void atlmix_show_ptd_header( unsigned char transfer_type, unsigned char index )
{
	unsigned short	p[6];
	unsigned short	target_address;
	
	target_address	= g_buffer_parameter[ transfer_type ].base_address + (g_buffer_parameter[ transfer_type ].payload_size + PTD_HEADER_SIZE) * index;
	
	read_direct_access( target_address, (unsigned char *)p, 12 );
	mprintf( WHITE, CONTINUE, "%d-%d : PTD-h 0x %04X %04X %04X %04X %04X %04X \r\n", transfer_type, index, *(p + 0), *(p + 1), *(p + 2), *(p + 3), *(p + 4), *(p + 5) );	
}

#endif


void atlmix_view_PTD( unsigned char index )
{
	atlmix_show_ptd_header( CTRL_BULK_TRANSFER, index );
}


void atlmix_display_open_slot( void )
{
	unsigned char	i;
	
	for ( i = 0; i < g_buffer_parameter[ CTRL_BULK_TRANSFER ].number_of_blocks; i++ )
		mprintf( YELLOW, CONTINUE, "%c", (g_atl_transfer_slot[ i ].lock == LOCK) ? 'L' : 'O' );

	mprintf( YELLOW, CONTINUE, "\r\n" );
}

#endif	//	DEBUG_CODE








void atlmix_buffer_parameter_initialize( buffer_info *bip )
{
	unsigned char	i;
	unsigned short	next_base_address;
	
	
	//	ISTL parameters
	
	next_base_address	= 0;
	
	g_buffer_parameter[ ISOCHRONOUS_TRANSFER ].base_address			= 0x0000;
	g_buffer_parameter[ ISOCHRONOUS_TRANSFER ].total_buffer_size	= ((bip->isochr).single_buffer_size) << 1;
	g_buffer_parameter[ ISOCHRONOUS_TRANSFER ].payload_size			= 0;
	g_buffer_parameter[ ISOCHRONOUS_TRANSFER ].number_of_blocks		= 0;

	next_base_address	+= g_buffer_parameter[ ISOCHRONOUS_TRANSFER ].total_buffer_size;	
	

	//	INTL parameters
	
	g_buffer_parameter[ INTERRUPT_TRANSFER   ].base_address			= next_base_address;
	g_buffer_parameter[ INTERRUPT_TRANSFER   ].total_buffer_size	= (PTD_HEADER_SIZE + (bip->interr).payload_size) * (bip->interr).n_of_blocks;
	g_buffer_parameter[ INTERRUPT_TRANSFER   ].payload_size			= (bip->interr).payload_size;
	g_buffer_parameter[ INTERRUPT_TRANSFER   ].number_of_blocks		= (bip->interr).n_of_blocks;

	next_base_address	+= g_buffer_parameter[ INTERRUPT_TRANSFER ].total_buffer_size;


	//	ATL parameters

#ifdef USE_PAIRED_PTD
	
	{
		unsigned char	c	= 0;
		
		for ( i = 0; i < 10; i++ )
			c	+= (((bip->aperio).payload_size >> i) & 0x1) ? 1 : 0;
		
		if (c != 1)
			mprintf( LIGHTRED, CONTINUE, "\r\n*** WARNING! ***\r\nFor the paired-PTD, the ATL payload size is needed to be \"(2^N) && (>=128)\".\r\n\r\n" );	
	}
	
#else

#endif

	g_buffer_parameter[ CTRL_BULK_TRANSFER   ].base_address			= next_base_address;
	g_buffer_parameter[ CTRL_BULK_TRANSFER   ].total_buffer_size	= 4096 - next_base_address;
	g_buffer_parameter[ CTRL_BULK_TRANSFER   ].payload_size			= (bip->aperio).payload_size;
	g_buffer_parameter[ CTRL_BULK_TRANSFER   ].number_of_blocks		= (4096 - next_base_address) / (PTD_HEADER_SIZE + (bip->aperio).payload_size);


	//	First block is reserved for ControlTransfer in this program

	g_buffer_parameter[ CONTROL_TRANSFER     ].base_address			= g_buffer_parameter[ CTRL_BULK_TRANSFER   ].base_address;
	g_buffer_parameter[ CONTROL_TRANSFER     ].total_buffer_size	= g_buffer_parameter[ CTRL_BULK_TRANSFER   ].payload_size + PTD_HEADER_SIZE;
	g_buffer_parameter[ CONTROL_TRANSFER     ].payload_size			= g_buffer_parameter[ CTRL_BULK_TRANSFER   ].payload_size;
	g_buffer_parameter[ CONTROL_TRANSFER     ].number_of_blocks		= 1;

	g_buffer_parameter[ BULK_TRANSFER        ].base_address			= g_buffer_parameter[ CTRL_BULK_TRANSFER   ].base_address + g_buffer_parameter[ CONTROL_TRANSFER     ].payload_size + PTD_HEADER_SIZE;
	g_buffer_parameter[ BULK_TRANSFER        ].total_buffer_size	= g_buffer_parameter[ CTRL_BULK_TRANSFER   ].total_buffer_size - g_buffer_parameter[ CONTROL_TRANSFER     ].total_buffer_size;
	g_buffer_parameter[ BULK_TRANSFER        ].payload_size			= g_buffer_parameter[ CTRL_BULK_TRANSFER   ].payload_size;
	g_buffer_parameter[ BULK_TRANSFER        ].number_of_blocks		= g_buffer_parameter[ CTRL_BULK_TRANSFER   ].number_of_blocks - 1;

	//	Buffer general
	write_register16( Com16_HcBufferStatus,                 0                                                                      );
	
	//	Buffer sizes
	write_register16( Com16_HcATLBufferSize,                g_buffer_parameter[ CTRL_BULK_TRANSFER   ].total_buffer_size             );
	write_register16( Com16_HcINTLBufferSize,               g_buffer_parameter[ INTERRUPT_TRANSFER   ].total_buffer_size             );
	write_register16( Com16_HcISTLBufferSize,              (g_buffer_parameter[ ISOCHRONOUS_TRANSFER ].total_buffer_size) >> 1       );	//	Set size of single buffer

	//	ISTL settings
	write_register16( Com16_HcISTLToggleRate,               0                                                                      );

	//	INTL settings
	write_register16( Com16_HcINTLBlkSize,					g_buffer_parameter[ INTERRUPT_TRANSFER   ].payload_size                  );
	write_register32( Com32_HcINTLPTDSkipMap,				0xFFFFFFFF                                                             );
	write_register32( Com32_HcINTLLastPTD,					0x1 << (g_buffer_parameter[ INTERRUPT_TRANSFER   ].number_of_blocks - 1) );
	
	//	ATL settings
	write_register16( Com16_HcATLBlkSize,					g_buffer_parameter[ CTRL_BULK_TRANSFER   ].payload_size                  );
	write_register32( Com32_HcATLPTDSkipMap,				0xFFFFFFFF                                                             );
	write_register32( Com32_HcATLLastPTD,					0x1 << (g_buffer_parameter[ CTRL_BULK_TRANSFER ].number_of_blocks - 1)   );
	write_register16( Com16_HcATLPTDDoneThresholdCount,     InitVal_HcATLPTDDoneThresholdCount                                     );
	write_register16( Com16_HcATLPTDDoneThresholdTimeOut,	InitVal_HcATLPTDDoneThresholdTimeOut                                   );

	if ( NULL == (g_atl_transfer_slot		= (transfer_instance *)malloc( sizeof( transfer_instance ) * g_buffer_parameter[ CTRL_BULK_TRANSFER ].number_of_blocks )) )
		mprintf( LIGHTRED, ABORT, "failed to make transfer instances for ATL slots for %u blocks", g_buffer_parameter[ BULK_TRANSFER ].number_of_blocks);

	if ( NULL == (g_intl_transfer_slot	= (transfer_instance *)malloc( sizeof( transfer_instance ) * g_buffer_parameter[ INTERRUPT_TRANSFER ].number_of_blocks )) )
		mprintf( LIGHTRED, ABORT, "failed to make transfer instances for INTL slots for %u blocks", g_buffer_parameter[ BULK_TRANSFER ].number_of_blocks);

	for ( i = 0; i < g_buffer_parameter[ CTRL_BULK_TRANSFER ].number_of_blocks; i++ )
	{
		atlmix_free_transfer_index( CTRL_BULK_TRANSFER, i );
		g_atl_transfer_slot[ i ].index			= i;
		g_atl_transfer_slot[ i ].transfer_type	= CTRL_BULK_TRANSFER;
	}
	
	for ( i = 0; i < g_buffer_parameter[ INTERRUPT_TRANSFER ].number_of_blocks; i++ )
	{
		atlmix_free_transfer_index( INTERRUPT_TRANSFER, i );
		g_intl_transfer_slot[ i ].index			= i;
		g_intl_transfer_slot[ i ].transfer_type	= INTERRUPT_TRANSFER;
	}
	
	for ( i = 0; i < 32; i++ )
		g_atl_slot_number_by_block[ i ]	= 0xFF;
	
}












