/*
**  WASABI-Hot! version 1.2c   
**
**
**      -- copyright (c) 2001-2004 by Philips Japan, Ltd. -- All rights reserved --
**
**
**      ** This code has been made to check/learn                          ** 
**      **                             the ISP1362/ISP1363 functionalities **
**      ** Release 06-Aug-2004                                             **
**
**      OKANO, Akifumi
**      
**		Application Laboratory, Mobile and Connectivity
**      Semiconductors Div, Philips Japan Ltd.
**      akifumi.okano@philips.com
**      +81-3-3740-4668 
*/

/*
**	HID-Keyboard/Mouse handling framework has been written by Hisashige Harashima
**		04-Mar-2002
*/



/****************************************************************************/
/*	includes																*/
/****************************************************************************/

#include		<stdio.h>
#include		<stdlib.h>
#include		<alloc.h>
#include		<limits.h>
#include		<mem.h>

#include		"_hc_core/atl_mix.h"
#include		"_hc_core/chap9.h"
#include		"_hc_core/dev_ep.h"
#include		"_hc_core/transfer.h"
#include		"_hc_core/dev_ep.h"

#include		"_hc_cls/cls_hndl.h"

#include 		"_otg/otg.h"

#include		"ui.h"
#include		"general.h"


/****************************************************************************/
/*	constants																*/
/****************************************************************************/

//	#define		Use_DESCROPTOR_SAVER		//	To use this option, make a folder named "log" in application folder
//	#define		Warn_unknown_DESCROPTOR

#define		BUFF_SIZE		UCHAR_MAX


/****************************************************************************/
/*	global vars																*/
/****************************************************************************/

device_instance		*device_list[ MAX_DEVICE_NUM ];
unsigned char		*g_config_desc_buffer_ptr		= NULL;
unsigned short		g_config_desc_buffer_length		= 0;


/****************************************************************************/
/*	function prototypes														*/
/****************************************************************************/

endpoint_info	*create_endpoint( device_instance *dvi_ptr, unsigned short index, unsigned short direction, unsigned short transfer_type, unsigned short max_packet_size );
unsigned short	get_string_descriptor_LANG_code( unsigned short *code, device_instance *dvi_ptr );
unsigned short	get_string_descriptor( unsigned short **unicode_str_ptr_ptr, device_instance *dvi_ptr, unsigned char index, unsigned short language_ID );
void			dispose_endpoint( device_instance *dvi_ptr, unsigned short index, unsigned char direction );
void			devep_device_list_linitialize( void );
unsigned char	device_new_address( void );


/****************************************************************************/
/*	function definitions													*/
/****************************************************************************/



/*************************************************
**
**  Create a device instance
**
*************************************************/

device_instance *devep_create_device( unsigned char speed, unsigned char topology_layer )
{
	device_instance		*dvi_ptr;
	unsigned char		buffer[ BUFF_SIZE ];
	unsigned short		rx_size;
	unsigned short		language_ID;
	int					i;
	unsigned char 		reserved_address;

//	mprintf( WHITE, CONTINUE, "creating device instance.\r\n" );
	
	/*	Check : Is new address available?  */
	
	if ( (0 == (reserved_address	= device_new_address())) )
	{
		mprintf( LIGHTRED, CONTINUE, "no available device address\r\n" );
		return ( NULL );
	}

	dvi_ptr		= devep_make_device_and_setup_ep0_connection( speed, topology_layer );

	if ( NULL == dvi_ptr )
	{
		mprintf( LIGHTRED, CONTINUE, "malloc failed @ newing device instance\r\n" );
		return ( NULL );
	}

	/*	Communication (enumeration) starts  */
	
	/********************************************/	
	/*	Get Descriptor [Device] 8 bytes			*/
	/*		to get MaxPacketSize of Endpoint 0	*/
	/********************************************/	
	//mprintf( WHITE, CONTINUE, "\r\nGet Descriptor [Device] 8 bytes    \r\n" );
	
	rx_size		= 8;
	
	if ( devep_std_request( buffer, dvi_ptr, std_Dv_request_GET_DESCRIPTOR, descriptor_type_DEVICE, 0x0000, &rx_size ) )
	{
		devep_dispose_device( &dvi_ptr );
		return ( NULL );
	}
	
	(dvi_ptr->epi_ptr[ EpOUT ][ 0 ])->max_packet_size	= ((std_dev_descriptor *)buffer)->bMaxPacketSize0;

	/********************************************/	
	/*	Set Address								*/
	/*		set new device address				*/
	/********************************************/	
	//mprintf( WHITE, CONTINUE, "\r\nSet Address    \r\n" );	
		
	dvi_ptr->address	= reserved_address;
	rx_size				= 0;
	
	if ( devep_std_request( buffer, dvi_ptr, std_Dv_request_SET_ADDRESS, (unsigned char)(dvi_ptr->address), 0x0000, &rx_size ) )
	{
		devep_dispose_device( &dvi_ptr );
		return ( NULL );
	}
	
	(dvi_ptr->epi_ptr[ EpOUT ][ 0 ])->dev_addr	= dvi_ptr->address;

	/********************************************/	
	/*	Get Descriptor [Device] 18 bytes		*/
	/*		to get full device descriptor		*/
	/********************************************/	
	//mprintf( WHITE, CONTINUE, "\r\nGet Descriptor [Device] 18 bytes    \r\n" );	
	
	rx_size				= 18;

	if ( devep_std_request( buffer, dvi_ptr, std_Dv_request_GET_DESCRIPTOR, descriptor_type_DEVICE, 0x0000, &rx_size ) )
	{
		devep_dispose_device( &dvi_ptr );
		return ( NULL );
	}
	
	memcpy( (unsigned char *)(&(dvi_ptr->dev_descriptor)), buffer, sizeof( std_dev_descriptor ) );


	/********************************************/	
	/*	Get Descriptor [String]					*/
	/*		If failed, returns a NULL			*/
	/********************************************/	
	//mprintf( WHITE, CONTINUE, "\r\nGet Descriptor [String]    \r\n" );


	if ( (dvi_ptr->dev_descriptor).iManufacturer | (dvi_ptr->dev_descriptor).iProduct | (dvi_ptr->dev_descriptor).iSerialNumber )
	{
		if ( get_string_descriptor_LANG_code( &language_ID, dvi_ptr ) )
		{
			devep_dispose_device( &dvi_ptr );
			return ( NULL );
		}

		if ( get_string_descriptor( &(dvi_ptr->string_manufacturer), dvi_ptr, (dvi_ptr->dev_descriptor).iManufacturer, language_ID ) )
		{
			devep_dispose_device( &dvi_ptr );
			return ( NULL );
		}
		
		if ( get_string_descriptor( &(dvi_ptr->string_product     ), dvi_ptr, (dvi_ptr->dev_descriptor).iProduct,      language_ID ) )
		{
			devep_dispose_device( &dvi_ptr );
			return ( NULL );
		}

		if ( get_string_descriptor( &(dvi_ptr->string_serialNum   ), dvi_ptr, (dvi_ptr->dev_descriptor).iSerialNumber, language_ID ) )
		{
			devep_dispose_device( &dvi_ptr );
			return ( NULL );
		}
	}

	
	/********************************************/	
	/*	Class procedure							*/
	/*		--------------------				*/
	/********************************************/	
	//mprintf( WHITE, CONTINUE, "\r\nClass procedure    \r\n" );	

	if ( NULL == (g_config_desc_buffer_ptr	= devep_get_configuration_and_interface( dvi_ptr )) )
	{
		devep_dispose_device( &dvi_ptr );
		return ( NULL );
	}

	if ( clshndl_initialization_method( dvi_ptr ) )
	{
		devep_dispose_device( &dvi_ptr );
		return ( NULL );
	}

	if ( NULL == dvi_ptr->class_instance_ptr )
	{
		//	This will set the device in default setting : Configuration=1, Interface=0 
	
		/********************************************/	
		/*	Get Descriptor [Cnfg] 					*/
		/*		--------------------				*/
		/********************************************/	

		if ( devep_regist_config_descriptors( dvi_ptr, 1, 0 ) )
		{
			devep_dispose_device( &dvi_ptr );
			return ( NULL );
		}
	}
	
	free( g_config_desc_buffer_ptr );
	
	g_config_desc_buffer_ptr		= NULL;
	g_config_desc_buffer_length		= 0;

	
	device_list[ reserved_address ]		= dvi_ptr;

//	ui_device_infomation_viewer( dvi_ptr );


	return ( dvi_ptr );
}


unsigned char *devep_get_configuration_and_interface( device_instance *dvi_ptr )
{
	unsigned char		*big_buffer_ptr;
	unsigned char		buffer[ 4 ];
	unsigned short		rx_size;


	/********************************************/	
	/*	Get Descriptor [Cnfg] 4 bytes			*/
	/*		to get the size of cnfg descriptor	*/
	/********************************************/	
	//mprintf( WHITE, CONTINUE, "\r\nGet Descriptor [Cnfg] 4 bytes    \r\n" );	
			
	rx_size				= 4;

	if ( devep_std_request( buffer, dvi_ptr, std_Dv_request_GET_DESCRIPTOR, descriptor_type_CONFIGURATION, 0x0000, &rx_size ) )
		return ( big_buffer_ptr );

	/********************************************/	
	/*	Get Descriptor [Cnfg] full size			*/
	/*		to get full cnfg... descriptor		*/
	/********************************************/	
	//mprintf( WHITE, CONTINUE, "\r\nGet Descriptor [Cnfg] full size    \r\n" );	
	
	rx_size				= 	((std_conf_descriptor *)buffer)->wTotalLength;

	if ( NULL == (big_buffer_ptr	= (unsigned char *)malloc( sizeof( unsigned char ) * rx_size )) )
		mprintf( LIGHTRED, ABORT, "malloc failed @ newing device instance (more size temp. buffer.\r\n" );

	if ( devep_std_request( big_buffer_ptr, dvi_ptr, std_Dv_request_GET_DESCRIPTOR, descriptor_type_CONFIGURATION, 0x0000, &rx_size ) )
	{
		free( big_buffer_ptr );
		big_buffer_ptr	= NULL;
		
		return ( big_buffer_ptr );
	}
	
	g_config_desc_buffer_length		= rx_size;
	
	/*	NOTE :	The CONFIGURATION descriptor is just 9 byte descriptor.	 */
	/*			INTERFACE, ENDPOINT descriptors will follow after this	 */

	/*	Set first config and IF descriptor for temporary use	*/

	memcpy( (unsigned char *)(&(dvi_ptr->conf_descriptor)),       (unsigned char *)(big_buffer_ptr + 0), sizeof( std_conf_descriptor )      );
	memcpy( (unsigned char *)(&(dvi_ptr->interfacef_descriptor)), (unsigned char *)(big_buffer_ptr + 9), sizeof( std_interface_descriptor ) );

	return ( big_buffer_ptr );
}





/*************************************************
**
**  Parse descriptors which was obtained by "get descriptor configuration"
**
*************************************************/


#define		descriptor_length( x )			*( x )
#define		descriptor_type( x )			*((x) + 1)

unsigned char devep_regist_config_descriptors( device_instance *dvi_ptr, unsigned char configuration_number, unsigned char interface_number )
{
	unsigned char				n_config		= 1;
	unsigned char				n_interface		= 0;
	unsigned char				*buf;
	unsigned char				*buf_end;
	unsigned char				desc_length;
	unsigned short				desc_type;
	std_endpoint_descriptor		tmp_ep_desc;
	endpoint_info				*epi_ptr;
	unsigned char				get_if_descriptors		= False;
	unsigned char				get_ep_descriptors		= False;
	
	if ( !configuration_number )	
		return ( 0 );

	{
		unsigned char	b;
		unsigned short	rx_size		= 0;
	
		if ( devep_std_request( &b, dvi_ptr, std_Dv_request_SET_CONFIGURATION, configuration_number, 0x0000, &rx_size ) )
			return ( 1 );
		
		if ( interface_number )
		{
			if ( devep_std_request( &b, dvi_ptr, std_Dv_request_SET_INTERFACE, interface_number, 0x0000, &rx_size ) )
				return ( 1 );
		}
	}
	

	buf			= g_config_desc_buffer_ptr;
	buf_end		= buf + g_config_desc_buffer_length;

	while ( buf < buf_end )
	{
		desc_length		= descriptor_length( buf );
		desc_type		= descriptor_type( buf ) << 8;	/*	See "_hc_core/chap9.h". Constants descriptor_type_XXX are 
															defined as upper byte in unsigned short				*/
		
		switch ( desc_type )
		{
			case descriptor_type_CONFIGURATION :
			
				if ( n_config == configuration_number )
				{
					memcpy( (unsigned char *)(&(dvi_ptr->conf_descriptor)), (unsigned char *)(buf), sizeof( std_conf_descriptor ) );
					get_if_descriptors	= True;
				}	
				else if ( n_config > configuration_number )
				{
					return ( 0 );
				}
				n_config++;
				
				break;
				
			case descriptor_type_INTERFACE :
			
				if ( get_if_descriptors )
				{
					if ( n_interface == interface_number )
					{
						memcpy( (unsigned char *)(&(dvi_ptr->interfacef_descriptor)), (unsigned char *)(buf), sizeof( std_interface_descriptor ) );
						get_ep_descriptors	= True;
					}	
					else if ( n_interface > interface_number )
					{
						return ( 0 );
					}
					n_interface++;
				}
				
				break;
				
			case descriptor_type_ENDPOINT :
			
				if ( get_ep_descriptors )
				{
					memcpy( (unsigned char *)(&(tmp_ep_desc)), (unsigned char *)(buf), sizeof( std_endpoint_descriptor ) );

					{
						unsigned char	direction;
						unsigned char	ep_address;
					
						direction		= EpDirection_bEndpointAddress( tmp_ep_desc.bEndpointAddress ) ? EpIN : EpOUT;
						ep_address		= EpNumber_bEndpointAddress( tmp_ep_desc.bEndpointAddress );
						
						if ( !(dvi_ptr->epi_ptr[ direction ][ ep_address ]) )
						{
							epi_ptr		= create_endpoint(	
															dvi_ptr, 
															ep_address,
															direction,
															Ep_TransferType_bmAttributes( tmp_ep_desc.bmAttributes ), 
															tmp_ep_desc.wMaxPacketSize
														);
															
							if ( NULL == epi_ptr )
								return ( 1 );
					
							epi_ptr->endpoint_descriptor					= tmp_ep_desc;		
							dvi_ptr->epi_ptr[ direction ][ ep_address ]		= epi_ptr;
						}
													
					}
				}
				
				break;
				
			case descriptor_type_OTG :
				
				if ( get_if_descriptors )
				{
					otg_handle_OTG_descriptor( (unsigned char *)(buf) );
				}
				
				break;
				
			default :
				break;
		}
		
		buf		+= desc_length;
	}
	
	
	return ( 0 );
}


/*************************************************
**
**  Get device instance pointer and setup EP0 connection
**
*************************************************/


device_instance *devep_make_device_and_setup_ep0_connection( unsigned char speed, unsigned char topology_layer )
{
	device_instance		*dvi_ptr;
	unsigned char		i;

	/*	Memory allocate for new device instance		*/

	if ( NULL == (dvi_ptr		= (device_instance *)malloc( sizeof( device_instance ) )) )
		return ( NULL );
	
	/*	Initialize members		*/
	
	dvi_ptr->address				= 0;
	dvi_ptr->speed					= speed;
	
	for ( i = 0; i < MAX_ENDPOINT_NUM; i++ )
		dvi_ptr->epi_ptr[ EpOUT ][ i ]	= NULL;

	for ( i = 0; i < MAX_ENDPOINT_NUM; i++ )
		dvi_ptr->epi_ptr[ EpIN  ][ i ]	= NULL;

	dvi_ptr->string_manufacturer	= NULL;
	dvi_ptr->string_product			= NULL;
	dvi_ptr->string_serialNum		= NULL;
	dvi_ptr->class_instance_ptr		= NULL;
	dvi_ptr->topology_layer			= topology_layer;

	dvi_ptr->epi_ptr[ EpOUT ][ 0 ]	= create_endpoint( dvi_ptr, 0, Ep_Direction_CONTROL_ENDPOINT, Ep_TransferType_Control, DEFAULT_BUFFER_SIZE );
	dvi_ptr->epi_ptr[ EpIN  ][ 0 ]	= dvi_ptr->epi_ptr[ EpOUT ][ 0 ];

	return ( dvi_ptr );
}


/*************************************************
**
**  Check a LANGUAGE code
**
*************************************************/



unsigned short get_string_descriptor_LANG_code( unsigned short *code, device_instance *dvi_ptr )
{
#define		LANG_code_request		0x00
	char				tmp_str[ 4 ];
	unsigned short		rx_size;
	unsigned short		err;
	
	rx_size		= 4;

	if ( 0 != (err	= devep_std_request( tmp_str, dvi_ptr, std_Dv_request_GET_DESCRIPTOR, descriptor_type_STRING | LANG_code_request, 0x0000, &rx_size )) )
		return ( err );
	
	 *code	= *(((unsigned short *)tmp_str) + 1);
	
	return ( 0 );
}



/*************************************************
**
**  Get a string descpiptor
**
*************************************************/


unsigned short get_string_descriptor( unsigned short **unicode_str_ptr_ptr, device_instance *dvi_ptr, unsigned char index, unsigned short language_ID )
{
	char				tmp_str[ 2 ];
	unsigned short		rx_size;
	unsigned short		err;

	if ( index == 0 )
	{
		*unicode_str_ptr_ptr	= NULL;
		return( 0 );
	}
	
	rx_size		= 2;

	if ( 0 != (err	= devep_std_request( tmp_str, dvi_ptr, std_Dv_request_GET_DESCRIPTOR, descriptor_type_STRING | index, language_ID, &rx_size )) )
		return ( err );
		
	rx_size		= ((std_string_descriptor *)tmp_str)->bLength;
	
	if ( NULL == (*unicode_str_ptr_ptr	= (unsigned short *)malloc( sizeof( unsigned char ) * rx_size )) )		//	Because the "rx_size" is given as a number of bytes.
		mprintf( LIGHTRED, ABORT, "error @ get_string_descriptor, malloc failed\r\n" );

	if ( 0 != (err	= devep_std_request( (unsigned char *)(*unicode_str_ptr_ptr), dvi_ptr, std_Dv_request_GET_DESCRIPTOR, descriptor_type_STRING | index, 0x0409, &rx_size )) )
		return ( err );

#if 0
{
unsigned char	*s;
int				i;

mprintf( LIGHTGREEN, CONTINUE, "str >> " );

for ( i = 1; i < rx_size>>1; i ++ )
mprintf( LIGHTGREEN, CONTINUE, "%c", *((*unicode_str_ptr_ptr) + i) & 0xFF );

mprintf( LIGHTGREEN, CONTINUE, " <<\r\n" );
}
#endif

	return ( 0 );
}



/*************************************************
**
**  Send/Receive standard request
**
*************************************************/


unsigned short devep_std_request( unsigned char *buff_ptr, device_instance *dvi_ptr, unsigned short request, unsigned short wValue, unsigned short wIndex, unsigned short *wLength )
{
	USB_Device_Request	req;
	unsigned short		err;

	req.bmRequestType			= reqtype( request );
	req.bRequest				= request( request );
	req.wValue					= wValue;
	req.wIndex					= wIndex;
	req.wLength					= *wLength;

	err			= transfer_control_transfer( &req, dvi_ptr, buff_ptr );

	if ( err & 0x8000 )						//	short data received
	{
		*wLength		= err & 0x7FFF;		//	In the case of no error, value is not updated.
		err 			= 0;				//	Clear error flag
	}
	
	return ( err );
}



/*************************************************
**
**  Dispose device instance
**
*************************************************/


void devep_dispose_device( device_instance **dvi_ptr_ptr )
{
	device_instance		*downstream_device_ptr;
	int					i;

	if ( (*dvi_ptr_ptr) == NULL )
		return;

	if ( NULL != ((*dvi_ptr_ptr)->class_instance_ptr) )
		clshndl_dispose_method( *dvi_ptr_ptr );
			
	device_list[ (*dvi_ptr_ptr)->address ]		= NULL;

	for ( i = 0; i < MAX_ENDPOINT_NUM; i++ )
		if ( NULL != (*dvi_ptr_ptr)->epi_ptr[ EpOUT ][ i ] )
			dispose_endpoint( *dvi_ptr_ptr, i, EpOUT );

	for ( i = 1; i < MAX_ENDPOINT_NUM; i++ )
		if ( NULL != (*dvi_ptr_ptr)->epi_ptr[ EpIN  ][ i ] )
			dispose_endpoint( *dvi_ptr_ptr, i, EpIN );

	if ( NULL != (*dvi_ptr_ptr)->string_manufacturer )
		free( (*dvi_ptr_ptr)->string_manufacturer );
		
	if ( NULL != (*dvi_ptr_ptr)->string_product )
		free( (*dvi_ptr_ptr)->string_product );
		
	if ( NULL != (*dvi_ptr_ptr)->string_serialNum )
		free( (*dvi_ptr_ptr)->string_serialNum );

	if ( NULL != (*dvi_ptr_ptr)->class_instance_ptr )
		free( (*dvi_ptr_ptr)->class_instance_ptr );

	free( *dvi_ptr_ptr );

	*dvi_ptr_ptr		= NULL;
}



/*************************************************
**
**  Cleate endpoint instance
**
*************************************************/


endpoint_info *create_endpoint( device_instance *dvi_ptr, unsigned short index, unsigned short direction, unsigned short transfer_type, unsigned short max_packet_size )
{
	endpoint_info	*epi_ptr;
	unsigned char	ep_direction;
	
	if ( NULL == (epi_ptr	= (endpoint_info *)malloc( sizeof( endpoint_info ) )) )
		mprintf( LIGHTRED, ABORT, "malloc failed @ epi_ptr malloc err\r\n" );

	if ( direction == Ep_Direction_CONTROL_ENDPOINT )	//	Endpoint instance slot 0-OUT is used for Control-EP
		direction	= EpOUT;

	dvi_ptr->epi_ptr[ direction ][ index ]	= epi_ptr;
	
	epi_ptr->max_packet_size	= max_packet_size;
	epi_ptr->dev_addr			= dvi_ptr->address;
	epi_ptr->ep_num				= index;
	epi_ptr->toggle				= 0;
	epi_ptr->speed				= dvi_ptr->speed;
	epi_ptr->direction			= direction;
	epi_ptr->transfer_type		= transfer_type;
	
	if ( transfer_type == Ep_TransferType_Isochronous )
		epi_ptr->tr_type		= Format_Iso;
	else
		epi_ptr->tr_type		= Format_Async;
	
	epi_ptr->transfer_instance	= NULL;

	return ( epi_ptr );
}


/*************************************************
**
**  Dispose endpoint instance
**
*************************************************/


void dispose_endpoint( device_instance *dvi_ptr, unsigned short index, unsigned char direction )
{
	endpoint_info_ptr	temp_epi_ptr;

	temp_epi_ptr	= dvi_ptr->epi_ptr[ direction ][ index ];

	if ( temp_epi_ptr->transfer_instance )
		atlmix_free_transfer( temp_epi_ptr->transfer_instance );
	
	free( temp_epi_ptr );

	temp_epi_ptr	= NULL;
}


void devep_device_list_linitialize( void )
{
	int		i;
	
	device_list[ 0 ]	= (device_instance *)1;
	device_list[ 1 ]	= (device_instance *)1;
	
	for ( i = 2; i < MAX_DEVICE_NUM; i++ )
		device_list[ i ]	= NULL;
}


unsigned char device_new_address( void )
{
	int		i;

	for ( i = 2; i < MAX_DEVICE_NUM; i++ )
		if ( !device_list[ i ] )
			return ( i );
	
	return ( 0 );
}


device_instance *devep_find_device( unsigned char address )
{
	return ( device_list[ address ] );
}


device_instance *devep_find_class_interface( unsigned char InterfaceClassID, unsigned char search_start_address )
{
	device_instance		*dviPtr;
	int					i;
	
	if ( search_start_address < 2 )
		search_start_address	= 2;
	
	for ( i = search_start_address; i < 128; i++ )
	{	
		if ( NULL == (dviPtr	= devep_find_device( i )) )
			continue;

		if ( devep_check_interface_class_ID( dviPtr, InterfaceClassID ) )
			return ( dviPtr );
	}
	
	return ( NULL );
}


unsigned char devep_check_interface_class_ID( device_instance *dviPtr, unsigned char InterfaceClassID )
{
	if ( InterfaceClassID == devep_get_interface_class_ID( dviPtr ) )
		return ( 1 );
	else
		return ( 0 );
}


unsigned char devep_get_interface_class_ID( device_instance *dviPtr )
{
	if ( 0x00 == devep_get_device_class_ID( dviPtr ) )
		return ( (dviPtr->interfacef_descriptor).bInterfaceClass );
	else
		return ( 0xFF );
}


unsigned char devep_get_device_class_ID( device_instance *dviPtr )
{
	return ( (dviPtr->dev_descriptor).bDeviceClass );
}


/* Modification by hara from here.*/

device_instance *find_protocol_interface( unsigned char InterfaceProtocolID, unsigned char search_start_address )
{
	device_instance		*dviPtr;
	int					i;
	
	if ( search_start_address < 2 )
		search_start_address	= 2;
	
	for ( i = search_start_address; i < 128; i++ )
	{	
		if ( NULL == (dviPtr	= devep_find_device( i )) )
			continue;

		if ( devep_check_interface_protocol_ID( dviPtr, InterfaceProtocolID ) )
			return ( dviPtr );
	}
	
	return ( NULL );
}


unsigned char devep_check_interface_protocol_ID( device_instance *dviPtr, unsigned char InterfaceProtocolID )
{
	if (InterfaceProtocolID == devep_get_interface_protocol_ID( dviPtr ) )
		return ( 1 );
	else
		return ( 0 );
}


unsigned char devep_get_interface_protocol_ID( device_instance *dviPtr )
{
	if ( 0x03 ==  devep_get_interface_class_ID( dviPtr ) )
		return ( (dviPtr->interfacef_descriptor).bInterfaceProtocol );
	else
		return ( 0xFF );
}



/* Modification by hara to here.*/




















