/*
**  WASABI-Hot! version 1.2c   
**
**
**      -- copyright (c) 2001-2004 by Philips Japan, Ltd. -- All rights reserved --
**
**
**      ** This code has been made to check/learn                          ** 
**      **                             the ISP1362/ISP1363 functionalities **
**      ** Release 06-Aug-2004                                             **
**
**      OKANO, Akifumi
**      
**		Application Laboratory, Mobile and Connectivity
**      Semiconductors Div, Philips Japan Ltd.
**      akifumi.okano@philips.com
**      +81-3-3740-4668 
*/


//	2002 02 27		Port connection disconnection process improved
//	2002 04 08		Device speed reading timing corrected  


/****************************************************************************/
/*	includes																*/
/****************************************************************************/

#include		<dos.h>
#include		<stdio.h>
#include		<conio.h>

#include		"_hc_core/atl_mix.h"
#include		"_hc_core/dev_ep.h"
#include		"_hc_core/isr.h"
#include		"_hc_core/port_ctl.h"
#include		"_hc_hw/hc_comm.h"
#include		"_hc_hw/hw_acces.h"
#include		"_hc_cls/cls_hndl.h"

#include 		"_otg/otg.h"

#include		"class_dr/hub/hub.h"

#include		"general.h"
#include		"beep.h"
#include		"ui.h"

/****************************************************************************/
/*	constants																*/
/****************************************************************************/

#define			ENUM_RETRY				3
#define			ENUM_RETRY_INTERVAL		1000L

#define			RESET_SECOVERY_TIME		100L		//	In the standard, it is 10mS. But some devices need more time to start enum.
													//	See section 7.1.7.3 in USB2.0 spec.

#define			LONGER_RESET						//	This is required for USB2.0 specification
													//	BusReset from root hub must have 50ms duration.

#define			ENUMERATION_DONE_BLINKING_INTERVAL					350L

#define			ROOT_HUB_ADDRESS									1


/****************************************************************************/
/*	global vars																*/
/****************************************************************************/

device_instance		*device_at_root_hub[ NUMBER_OF_ROOT_HUB_PORT ]		=	{
																				NULL,
																				NULL
																		};
																		
hub_instance		gp_root_hub_instance								=	{
																				NULL,														//	Pointer to device instance
																				{	0x09, 0x29, 0x02, 0x09, 0x00, 0x32, 0x00, 0x00, 0x00 },	//	Dummy hub descriptor
																				2,															//	Number of downstream ports
																				{	NULL, NULL, NULL, NULL, NULL, NULL, NULL },				//	Pointer to device attached on downstream ports			
																				0xFF,														//	Transfer index for interrupt transfer
																				NULL,														//	Pointer to transfer instance
																				0															//	Hub and Port Statua Change Bitmap
																		};
																		
hub_event_flag		gp_hub_event										=	{
																				0,
																				0x00
																		};

unsigned short HcRhPortStatusComm[2]									=	{ 
																				Com32_HcRhPortStatus1,
																				Com32_HcRhPortStatus2
																			};

unsigned char		gp_fast_enumeration								= Disabled;
unsigned char		gp_silent_enumeration							= Disabled;
unsigned char		gp_warning_message								= Disabled;


/****************************************************************************/
/*	function prototypes														*/
/****************************************************************************/

void				isr_flag_handler( void );
void				root_hub_event( hub_instance *hi_ptr );
void				hub_event_handler( hub_event_flag *h_ev_ptr );
unsigned long		read_hub_port_status( unsigned char address, unsigned char port );
void				clear_status_change_bits( unsigned char address, unsigned char port, unsigned long status_bits );
device_instance		*connect_a_device( unsigned char address, unsigned char port, unsigned char topology_layer );
void				disconnect_a_device( device_instance **dvi_ptr_ptr );
unsigned char		port_reset( unsigned char address, unsigned char port, unsigned short reset_recovery_time );
device_instance		*start_a_device( unsigned char speed, unsigned char topology_layer );
void				message_of_enum_done( device_instance *dvi_ptr );


/****************************************************************************/
/*	function definitions													*/
/****************************************************************************/

void portctl_isr_flag_routine_dspatch( void )
{
	isr_flag_handler();
	
	hub_event_handler( &gp_hub_event );
}


void isr_flag_handler( void )
{
	unsigned short	flag_mask;	/*  this mask shold be set to */

	flag_mask	= SOFISTLInt | ATL_IRQ | AllEOTInterrupt;

	if ( !(gp_isr_flag & ~flag_mask) )
		return;

	root_hub_event( &gp_root_hub_instance );
}


void root_hub_event( hub_instance *dummy_for_func_compatibility )
{
	static unsigned char	rh_Vbus_status[ 2 ]		= { 0, 0 };		//	##	version1.2c
	unsigned long			rh_status;								//	##	version1.2c
	unsigned char			i;										//	##	version1.2c

	dummy_for_func_compatibility	= dummy_for_func_compatibility;	//	This line for erase warning

	if ( gp_isr_flag & RHSC )
		gp_isr_flag	&= ~RHSC;
	else
		return;

	gp_hub_event.port_status_change_bit_map	= 0x0;

	for ( i = 0; i < 2; i++ )
	{
		rh_status	= read_register32( HcRhPortStatusComm[ i ] );

#ifdef	PORT1_AS_HOST_ONLY
#else
		if ( i == 0 )
		{
			if ( !OTG_host_operation_enabled() )
				continue;
			else if ( (rh_status & 0x1) && (gp_root_hub_instance.portctl_device_on_port)[ 0 ] )
				continue;
		}
#endif

		//	##	version1.2c	(from here)

		if ( (rh_status & 0x001F0000) || (rh_Vbus_status[ i ] && (~rh_status & 0x100)) )	//	##	version1.2c
		{
			//	If this event is happened root-hub, over-ride the any hub events. 
			//	Root-hub event has highest priority
			
			if ( rh_status & 0x00080000 )
				ui_over_current_detected_message( i );

			gp_hub_event.hub_address				 = 1;
			gp_hub_event.port_status_change_bit_map	|= (0x1 << (i + 1));
			gp_hub_event.topology_layer				 = 0;
		}
		
		rh_Vbus_status[ i ]		= (rh_status & 0x100) ? True : False;	//	##	version1.2c

		//	##	version1.2c	(to here)

	}
}


void hub_event_handler( hub_event_flag *h_ev_ptr )
{
	device_instance				*dvi_ptr;
	hub_instance				*hi_ptr;
	unsigned long				h_status;
	unsigned char				port_num;
	int							i;

	if ( h_ev_ptr->hub_address == 0 )
		return;
		
	if ( NULL == (dvi_ptr	= devep_find_device( h_ev_ptr->hub_address )) )
		return;
	
	
	//	Set target HUB
	
	if ( h_ev_ptr->hub_address == ROOT_HUB_ADDRESS )
	{
		hi_ptr	= &gp_root_hub_instance;				//	If the target is a root HUB
	}
	else
	{
		hi_ptr	= hub_instance_ptr( dvi_ptr );		//	If the target is a device HUB
	}
	
	
	//	Port processing for target HUB

	for ( i = 0; i < hi_ptr->number_of_ports; i++ )
	{
		port_num	= i + 1;
		
		if ( !(h_ev_ptr->port_status_change_bit_map & (0x1 << port_num)) )
			continue;
			
		h_status		= read_hub_port_status( h_ev_ptr->hub_address, port_num );

		if ( h_status & 0x001F0000 )
		{
			//	This function clears register bits for hub.
			//	It clears interrupt transfer data body from device-hub
			clear_status_change_bits( h_ev_ptr->hub_address, port_num, h_status & 0x001F0000 );
		}


		//	##	version1.2c	(from here)

		if ( (h_status & 0x00010000) && ( h_status & 0x00000001) )		//	Device connected!
		{
			if ( NULL == hi_ptr->portctl_device_on_port[ i ] )

				hi_ptr->portctl_device_on_port[ i ]		= connect_a_device( h_ev_ptr->hub_address, i + 1, h_ev_ptr->topology_layer + 1 );


		}
		else if ( ((h_status & 0x00010000) && (~h_status & 0x00000001)) || ((hi_ptr->portctl_device_on_port[ i ] != NULL) && (~h_status & 0x00000001)) )
					//	Device disconnected!                            or     Vbus turned-off while the device instance exist
		{
			disconnect_a_device( &(hi_ptr->portctl_device_on_port[ i ]) );
		}
		//	##	version1.2c	(from here)
	}

	h_ev_ptr->hub_address	= 0;

#if 0

	if ( h_status & 0x00010000 )
		ui_show_topology();
#endif
}


unsigned long read_hub_port_status( unsigned char address, unsigned char port )
{
	if ( address == ROOT_HUB_ADDRESS )
		return ( read_register32( HcRhPortStatusComm[ port - 1 ] ) );
	else
		return ( hub_read_hub_device_port_status( devep_find_device( address ), port ) );
}


void clear_status_change_bits( unsigned char address, unsigned char port, unsigned long status_bits )
{
	int		i;
	
	if ( address == ROOT_HUB_ADDRESS )
		write_register32( HcRhPortStatusComm[ port - 1 ], status_bits );
	else
	{
		for ( i = 16; i <=20; i++ )
		{
			if ( status_bits & (0x1L << i) )
				hub_clear_status_change_bits_in_hub(  devep_find_device( address ), port, i );
		}
	}
}


device_instance *connect_a_device( unsigned char address, unsigned char port, unsigned char topology_layer )
{
	device_instance		*dvi_ptr;
	unsigned char		speed;
	unsigned char		retry;
	unsigned char		i;
	unsigned char		fast_enumeration_flag_to_remember;
	
	dvi_ptr		= NULL;

#if 0
	if ( address == ROOT_HUB_ADDRESS )
		mprintf( LIGHTGRAY, CONTINUE, "\r\ndetected a connection @root_hub" );
	else
		mprintf( LIGHTGRAY, CONTINUE, "\r\ndetected a connection @hub(addr=%d)", address );
#endif


#if 0
	if ( gp_fast_enumeration )
		retry	= 1;
	else
		retry	= ENUM_RETRY;
#else
		retry	= ENUM_RETRY;
#endif

	fast_enumeration_flag_to_remember	= gp_fast_enumeration;

	if ( (address == ROOT_HUB_ADDRESS) && (port == 1) && OTG_host_operation_enabled() )
		gp_fast_enumeration		= 1;

	for ( i = 0; i < retry; i++ )
	{
		speed	= port_reset( address, port, RESET_SECOVERY_TIME );		//	reset a port on ( address, port );

		dvi_ptr		= start_a_device( speed, topology_layer );			//	start a device!

		if ( dvi_ptr != NULL )
			break;
		
//		read_register32( HcRhPortStatusComm[ port - 1 ] & 0x00000001 );	//	Disable port

		if ( gp_warning_message )
			mprintf( YELLOW, CONTINUE, "*** Enumeration RETRY wait #%d\r", i );

		wait_ms( ENUM_RETRY_INTERVAL );
	}
	
	gp_fast_enumeration	= fast_enumeration_flag_to_remember;

	if ( i == retry )
		mprintf( YELLOW, CONTINUE, "\r\n" );

	return ( dvi_ptr );	
}


void disconnect_a_device( device_instance **dvi_ptr_ptr )
{
	unsigned char	address;

	mprintf( LIGHTGRAY, CONTINUE, "\r\ndetected a device disconnection\r\n" );

	if ( *dvi_ptr_ptr == NULL )
	{
		mprintf( LIGHTGRAY, CONTINUE, "un-enumerated device removed.\r\n" );
		return;
	}
	
	address		= (*dvi_ptr_ptr)->address;
	devep_dispose_device( dvi_ptr_ptr );
	
	*dvi_ptr_ptr	= NULL;

	mprintf( WHITE, CONTINUE, "device (address=%d) disposed\r\n", address );
	
	if ( !gp_silent_enumeration )
		beep( 660, 0.20, DECAY, POLY );
}

#define		OTG__SESSION_BASED

unsigned char port_reset( unsigned char address, unsigned char port, unsigned short reset_recovery_time )
{
	unsigned long		h_status;

#ifdef	OTG__SESSION_BASED

	if ( (address == ROOT_HUB_ADDRESS) && (port == 1) && OTG_host_operation_enabled() )
		;																			//	If the local is OTG host, no need to wait.
	else
		wait_ms( 100L );															//	Need to wait before reset to handle bounce

#else

	wait_ms( 100L );																//	Need to wait before reset to handle bounce

#endif


	if ( address == ROOT_HUB_ADDRESS )
	{
		write_register32( HcRhPortStatusComm[ port - 1 ], 0x00000010 );				//	Set port reset

#ifdef LONGER_RESET
		{
			unsigned char	i;
			
			for ( i = 0; i < 41; i++ )
			{
				write_register32( HcRhPortStatusComm[ port - 1 ], 0x00000010 );				//	Set port reset
				wait_ms( 1L );
			}

		}
#else
#endif

		while ( read_register32( HcRhPortStatusComm[ port - 1 ] ) & 0x00000010 )	//	Polling port reset status (waiting for finishing port reset)
			wait_ms( 5L );

		h_status	= read_register32( HcRhPortStatusComm[ port - 1 ] );

		wait_ms( reset_recovery_time );												//	Wait reset recovery time
	}
	else
	{
		hub_reset_hub_device_port( devep_find_device( address ), port );
	
		h_status	= hub_read_hub_device_port_status( devep_find_device( address ), port );

		wait_ms( reset_recovery_time );												//	Wait reset recovery time

		if ( 0x3 == h_status )
			mprintf( LIGHTGRAY, CONTINUE, "\r\nport enabled @hub(addr=%d) port%d\r\n", address, port );
	}

	return ( (h_status >> 9) & 0x1 );
}


device_instance *start_a_device( unsigned char speed, unsigned char topology_layer )
{
	device_instance		*dvi_ptr;
	unsigned char		dv_class_code;
	unsigned char		if_class_code;
	
	if ( NULL == (dvi_ptr	= devep_create_device( speed, topology_layer )) )		//	Do enumeration
	{
		if ( gp_warning_message )
			mprintf( YELLOW, CONTINUE, "\r\nWarning : device may have problem for enumeration process\r\n" );
			
		return ( NULL );
	}
	
	message_of_enum_done( dvi_ptr );

	return ( dvi_ptr );
}


void message_of_enum_done( device_instance *dvi_ptr )
{
	int		i,
			j;

	if ( !gp_silent_enumeration )
	{
		if ( devep_check_interface_class_ID( dvi_ptr, AUDIO_CLASS_INTERFACE ) )
			beep( 886.00, 1.20, DECAY, POLY );
		else
			beep( 443.00, 0.50, DECAY, POLY );
	}

	if ( gp_fast_enumeration )
	{
		mprintf( WHITE, CONTINUE, "device enumerated (address=%d assigned.)", dvi_ptr->address );
		mprintf( DARKGRAY, CONTINUE, "     vid=0x%04X pid=0x%04X\r", (dvi_ptr->dev_descriptor).idVendor, (dvi_ptr->dev_descriptor).idProduct );
//		mprintf( WHITE, CONTINUE, "device enumerated (address=%d assigned.)\r", dvi_ptr->address );
	}
	else
	{
		for ( j = 0; j < 5; j++ )
		{
//			mprintf( (j & 0x1) ? BLACK : WHITE, CONTINUE, "device enumerated (address=%d assigned. VID=0x%04X PID=0x%04X )\r", dvi_ptr->address, (dvi_ptr->dev_descriptor).idVendor, (dvi_ptr->dev_descriptor).idProduct );
			mprintf( (j & 0x1) ? BLACK : WHITE, CONTINUE, "device enumerated (address=%d assigned.)\r", dvi_ptr->address );
			wait_ms( ENUMERATION_DONE_BLINKING_INTERVAL );
		}
	}

	mprintf( WHITE, CONTINUE, "\r\n" );

	ui_show_device_string_descriptors( dvi_ptr );

//	ui_device_infomation_viewer( dvi_ptr );
}


device_instance *portctl_device_on_port( unsigned char port_num )
{
	return ( device_at_root_hub[ port_num - 1 ] );
}


void portctl_force_rh_port_reconnect( void )
{
	write_register32( Com32_HcRhPortStatus1, 0x1 );
	write_register32( Com32_HcRhPortStatus2, 0x1 );

	if ( (gp_root_hub_instance.portctl_device_on_port)[ 0 ] != NULL )
		disconnect_a_device( &((gp_root_hub_instance.portctl_device_on_port)[ 0 ]) );

	if ( (gp_root_hub_instance.portctl_device_on_port)[ 1 ] != NULL )
		disconnect_a_device( &((gp_root_hub_instance.portctl_device_on_port)[ 1 ]) );

	if ( (read_register32( Com32_HcRhPortStatus1 ) & 0x1) )
		(gp_root_hub_instance.portctl_device_on_port)[ 0 ]		= connect_a_device( 1, 1, 1 );
		
	if ( (read_register32( Com32_HcRhPortStatus2 ) & 0x1) )
		(gp_root_hub_instance.portctl_device_on_port)[ 1 ]		= connect_a_device( 1, 2, 1 );
}


unsigned char portctl_check_OTG_port( unsigned char bus_request, unsigned char full_enum )
{
	device_instance		*dvi_ptr		= NULL;
	unsigned long		port_status;

	write_register32( Com32_HcRhPortStatus1, 0x001F0000 );	//	Clearing ststus change bit

	if ( (gp_root_hub_instance.portctl_device_on_port)[ 0 ] != NULL )
		return ( False );
	
	port_status		= read_register32( Com32_HcRhPortStatus1 );

	if ( (port_status & 0x00000001) )						//	Device connected!
	{
	
#if 0
		if ( (!bus_request) && (!full_enum) )
					= otg_host_process_for_HNP();			//	Try to access as remote=DRD
			(gp_root_hub_instance.portctl_device_on_port)[ 0 ]		= connect_a_device( 1, 0 + 1, 0 + 1 );
		else
#endif


		if ( (!bus_request) && (!full_enum) )
			dvi_ptr		= otg_host_process_for_HNP();			//	Try to access as remote=DRD
		
		if ( dvi_ptr == NULL )
			dvi_ptr		= connect_a_device( 1, 0 + 1, 0 + 1 );
	}
	
	(gp_root_hub_instance.portctl_device_on_port)[ 0 ]	= dvi_ptr;

	return ( dvi_ptr ? True : False );
}


void portctl_disconnect_a_device_on_OTG_port( void )
{
	if ( (gp_root_hub_instance.portctl_device_on_port)[ 0 ] != NULL )
		disconnect_a_device( &((gp_root_hub_instance.portctl_device_on_port)[ 0 ]) );
	
	write_register32( Com32_HcRhPortStatus1, 0x00000001 );	//	Disabling OTG port (port1)
	write_register32( Com32_HcRhPortStatus1, 0x001F0000 );	//	Clearing ststus change bit
}


device_instance *portctl_get_OTG_port_device_instance( void )
{
	return ( (gp_root_hub_instance.portctl_device_on_port)[ 0 ] );
}


void portctl_OTG_port_enable_disable( unsigned char enable )
{
	unsigned long	v;
	unsigned char	state;
	

	state	= (read_register32( Com32_HcRhPortStatus1 ) & 0x00000002) ? True       : False;
	v		= enable                                                  ? 0x00000182 : 0x00000181;

	if ( state ^ enable )
		write_register32( Com32_HcRhPortStatus1, v );
}


unsigned char portctl_reset_OTG_port( void )
{
	return ( port_reset( ROOT_HUB_ADDRESS, 1, 10 ) ) ;
}


void portctl_Vbus_control_for_port_1_or_2( unsigned char power, unsigned char port_num )
{
	if ( power )
		write_register32( HcRhPortStatusComm[ port_num - 1 ], 0x00000100 );
	else
		write_register32( HcRhPortStatusComm[ port_num - 1 ], 0x00000200 );
}
























