/*
**  WASABI-Hot! version 1.2c    (Sing sub-unit)
**
**
**      -- copyright (c) 2001-2004 by Philips Japan, Ltd. -- All rights reserved --
**
**
**      ** This code has been made to check/learn                          ** 
**      **                             the ISP1362/ISP1363 functionalities **
**      ** Release 06-Aug-2004                                             **
**
**      OKANO, Akifumi
**      
**		Application Laboratory, Mobile and Connectivity
**      Semiconductors Div, Philips Japan Ltd.
**      akifumi.okano@philips.com
**      +81-3-3740-4668 
*/


/****************************************************************************/
/*	includes																*/
/****************************************************************************/

#include	<stdio.h>
#include	<math.h>
#include	<limits.h>

#include	"_hc_core/dev_ep.h"
#include	"ui.h"

#include	"_hc_cls/cls_hndl.h"

#include	"class_dr/audio/sing.h"

#include	"beep.h"


/****************************************************************************/
/*	constants																*/
/****************************************************************************/

#define		PI				3.1415926535


/****************************************************************************/
/*	global vars																*/
/****************************************************************************/

unsigned long		g_beep_sample_count;
unsigned long		g_duration;
unsigned char		g_decay_sw;

unsigned char		g_pippi_rate;

double				g_freq_factor_L;
double				g_freq_factor_R;
double				g_freq_factor_C;


/****************************************************************************/
/*	function prototypes														*/
/****************************************************************************/

void				make_a_sound( double frequency, double duration, unsigned char decay, unsigned char poly, unsigned short (*buff_fill_func)( short *buffer_ptr ) );
unsigned short 		buffer_fill_pi_pi( short *buffer_ptr );
unsigned short		buffer_fill_beep_sound( short *buffer_ptr );
unsigned short		buffer_fill_warn( short *buffer_ptr );


/****************************************************************************/
/*	function definitions													*/
/****************************************************************************/

void beep( double frequency, double duration, unsigned char decay, unsigned char poly )
{
	make_a_sound( frequency, duration, decay, poly, buffer_fill_beep_sound );
}


void beep_pippi( double frequency, double duration, unsigned char rate )
{
#if 1
	g_pippi_rate	= rate + 9;
	make_a_sound( frequency, duration, SUSTAIN, POLY, buffer_fill_pi_pi );
#else
	frequency	= frequency;
	duration	= duration;
	rate		= rate;
#endif
}


void beep_warn( double frequency, double duration )
{
	make_a_sound( frequency, duration, SUSTAIN, POLY, buffer_fill_warn );
}


void make_a_sound( double frequency, double duration, unsigned char decay, unsigned char poly, unsigned short (*buff_fill_func)( short *buffer_ptr ) )
{
	device_instance		*dvi_ptr;

	if ( is_audio_active() )
		return;


	/*								*/
	/*  Finding USB audio device	*/
	/*								*/
	
	if ( NULL == (dvi_ptr		= devep_find_class_interface( AUDIO_CLASS_INTERFACE, 0 )) )
		return ;

	g_beep_sample_count		= 0;
	g_duration				= (unsigned long)(duration * 44100);
	g_decay_sw				= decay;
	
	g_freq_factor_L			= frequency * ((2.00 * PI) / 44100.00);
	
	if ( poly == MONO )
	{
//		g_freq_factor_C			= g_freq_factor_L * 0.00;
		g_freq_factor_R			= g_freq_factor_L;
	}
	else
	{
//		g_freq_factor_C			= g_freq_factor_L * 1.3333333333;
		g_freq_factor_R			= g_freq_factor_L * 1.5000000000;
	}

	/*								*/
	/*	Initialize Audio functions	*/
	/*								*/
	
	if ( 0 != audio_initialize( dvi_ptr, NULL, buff_fill_func, NULL ) )
		return;
		
		
	/*								*/
	/*	Start to play Audio 		*/
	/*								*/

	audio_start();
}


unsigned short buffer_fill_beep_sound( short *buffer_ptr )
{
	short				*bf;
	short				sample_L;
	short				sample_C;
	short				sample_R;
	double				amp;
	unsigned int		i			= 0;

	bf		= (short *)buffer_ptr;

	if ( g_decay_sw )
	{
		amp		    = 0.30 * ((double)(g_duration - g_beep_sample_count) / (double)g_duration);			//	linear decay
	}
	else
	{
		amp		    = 0.30;
	}
	
	amp		   *= (double)SHRT_MAX;
	
	while ( (i < (MEMORY_FILLING_SIZE_FROM_FILE >> 1)) )
	{
		if ( g_duration <= g_beep_sample_count )
			break;

		sample_L			= (short)( sin( (double)g_beep_sample_count * g_freq_factor_L ) * amp);
		sample_R			= (short)( sin( (double)g_beep_sample_count * g_freq_factor_R ) * amp);

		*bf++	= sample_L;
		*bf++	= sample_R;
		
		g_beep_sample_count++;
		i++;
	}
	
	return ( i << 1 );
}



unsigned short buffer_fill_pi_pi( short *buffer_ptr )
{
	short				*bf;
	short				sample_L;
	short				sample_R;
	double				amp;
	unsigned int		i			= 0;

	bf		= (short *)buffer_ptr;

	amp		   = (double)(SHRT_MAX >> 1);
	
	while ( (i < (MEMORY_FILLING_SIZE_FROM_FILE >> 1)) )
	{
		if ( g_duration <= g_beep_sample_count )
			break;
		
		if ( (g_beep_sample_count >> g_pippi_rate) & 0x1 )
		{
			sample_L			= 0;
			sample_R			= (short)( sin( (double)g_beep_sample_count * g_freq_factor_L ) * amp);
		}
		else
		{
			sample_L			= (short)( sin( (double)g_beep_sample_count * g_freq_factor_R ) * amp);
			sample_R			= 0;
		}
		
		*bf++	= sample_L;
		*bf++	= sample_R;
		
		g_beep_sample_count++;
		i++;
	}
	
	return ( i << 1 );
}



unsigned short buffer_fill_warn( short *buffer_ptr )
{
	short				*bf;
	short				sample_L;
	short				sample_R;
	double				amp;
	unsigned int		i			= 0;
	
	double				local_freq_L;
	double				local_freq_R;

	
	bf		= (short *)buffer_ptr;

	amp		    = 0.90;	
	amp		   *= (double)SHRT_MAX;
	

	while ( (i < (MEMORY_FILLING_SIZE_FROM_FILE >> 1)) )
	{
		if ( g_duration <= g_beep_sample_count )
			break;

		local_freq_L		= g_freq_factor_L * ((double)g_beep_sample_count / (double)g_duration + 1.00);
		local_freq_R		= local_freq_L * 1.50;
		sample_L			= (short)( sin( (double)g_beep_sample_count * local_freq_L ) * amp);
		sample_R			= (short)( sin( (double)g_beep_sample_count * local_freq_R ) * amp);
		*bf++	= sample_L;
		*bf++	= sample_R;
		
		g_beep_sample_count++;
		i++;
	}
	
	return ( i << 1 );
}



