/*
**  WASABI-Hot! version 1.2c   
**
**
**      -- copyright (c) 2001-2004 by Philips Japan, Ltd. -- All rights reserved --
**
**
**      ** This code has been made to check/learn                          ** 
**      **                             the ISP1362/ISP1363 functionalities **
**      ** Release 06-Aug-2004                                             **
**
**      OKANO, Akifumi
**      
**		Application Laboratory, Mobile and Connectivity
**      Semiconductors Div, Philips Japan Ltd.
**      akifumi.okano@philips.com
**      +81-3-3740-4668
*/


/****************************************************************************/
/*	includes																*/
/****************************************************************************/


#include		<alloc.h>

#include		"class_dr/storage/storage.h"
#include		"class_dr/storage/fs.h"
#include		"ui.h"
#include		"general.h"


/****************************************************************************/
/*	constants																*/
/****************************************************************************/



//	directory accessing mode
#define			DIR_ACCESS_READ				0
#define			DIR_ACCESS_MODIFY			1
#define			DIR_ACCESS_APPEND_NEW		2
#define			DIR_ACCESS_APPEND_OW		3

#define			FAT_CACHE_NOT_VALID			0xFFFFFFFF
#define			DIR_CACHE_NOT_VALID			0xFFFFFFFF

#define			NO_DIR_MODIFY				0
#define			OK_DIR_MODIFY				1

#define			NO_MAKE_NEW_DIR_ENT			0
#define			OK_MAKE_NEW_DIR_ENT			1

#define			FILE_SYSTEM_FAT16			0
#define			FILE_SYSTEM_FAT12			1
#define			FILE_SYSTEM_UNKNOWN			0xFF


/****************************************************************************/
/*	global vars																*/
/****************************************************************************/

drive_param		gp_drive_parameter;
volume_param	g_volume_parameter;

unsigned char	*g_fs_scratch_buffer_ptr;
unsigned short	g_fs_scratch_buffer_size;
unsigned char	g_num_of_fs_client			= 0;
unsigned char	g_recent_error;

#if BIG_ENDIAN

directory_entry	g_root_dumy		=	{ "ROOTROOT", "DIR", 0x10, {0,0,0,0,0,0,0,0,0,0}, 0, 0, 
												//	ROOT_DIRECTORY_CLUSTER, 
												(ROOT_DIRECTORY_CLUSTER << 8) | (ROOT_DIRECTORY_CLUSTER >> 8), 
									0 };

#else

directory_entry	g_root_dumy		=	{ "ROOTROOT", "DIR", 0x10, {0,0,0,0,0,0,0,0,0,0}, 0, 0, ROOT_DIRECTORY_CLUSTER, 0 };

#endif


/****************************************************************************/
/*	function prototypes														*/
/****************************************************************************/

static unsigned short	fs_read_large( fs_FILE *fp, unsigned char *buffer_ptr, unsigned long clusters );
static unsigned short	fs_read_small( fs_FILE *fp, unsigned char *buffer_ptr, unsigned short length );
static unsigned short	fs_write_large( fs_FILE *fp, unsigned char *buffer_ptr, unsigned long clusters );
static unsigned short	fs_write_small( fs_FILE *fp, unsigned char *buffer_ptr, unsigned short length );

static unsigned char	make_directory_entry_fp( fs_FILE *fp );
static unsigned char	make_directory_entry( char *full_path, unsigned short parent_cluster, unsigned short *index_ptr, directory_entry *dir_ptr, unsigned char make_dir );
static unsigned char	update_directory_entry( fs_FILE *fp );
static unsigned char	remove_this_entry( unsigned short base_cluster, unsigned short index, unsigned char flags, unsigned char *ep, unsigned short depth );
static unsigned short	count_free_cluster( void );
static unsigned short	find_free_cluster( unsigned short start_cluster );
static unsigned char	clear_FAT_chain( unsigned short cluster );
static unsigned char	valid_cluster( unsigned short cluster, volume_param *vpp );
static unsigned short	read_FAT( unsigned short current_cluster );
static unsigned short	write_FAT( unsigned short cluster, unsigned short value );
static unsigned short	flush_FAT_cache( volume_param *vpp );
static unsigned short	flush_FAT16_cache( volume_param *vpp );
static unsigned short	flush_FAT12_cache( volume_param *vpp );
static unsigned short	access_FAT16( unsigned char mode, unsigned short cluster, unsigned short value );
static unsigned short	access_FAT12( unsigned char mode, unsigned short cluster, unsigned short value );
static unsigned long	cluster_to_sector( unsigned short cluster );
static directory_entry	*find_entry( char *path );
static directory_entry	*find_file_entry( char *path, unsigned short *dir_cluster_ptr, unsigned short *index_ptr );
static directory_entry	*find_directory_index( unsigned short *found_index, unsigned short base_cluster, char *name );
static directory_entry	*find_free_directory_entry( unsigned short *found_index, unsigned short base_cluster, char *name );
static directory_entry	*get_directory_entry( unsigned short base_cluster, unsigned short index, unsigned char new_entry_ok );
static void				write_back_directory_entry( void );

static void				set_time_for_directory_entry( directory_entry *d );
static unsigned char	format_file_name_directory_entry( char *target_base, char *source );
static unsigned char	str_match( char *a, char *b, unsigned char n );
static void				make_full_path( char *target, char *relative, char *current );
static char				*reform_path( char *source );
static void				separate_directory_and_file( char *directory_path, char *file_name, char *source );
static unsigned short	first_char_position( char *path, char mc );
static unsigned short	last_char_position( char *path, char mc );


/****************************************************************************/
/*	function definitions													*/
/****************************************************************************/


/*
==============================================================================
==============================================================================
====	Public functions for file system access
====		Functions for APPLICATION side
==============================================================================
==============================================================================
*/

//////////
//////////	"cd" change directory
//////////
unsigned char fs_cd( char *path )
{
	char				full_path[ MAX_PATH_STRING_LENGTH ];
	directory_entry		*dp;
	volume_param		*vpp;
	
	vpp				= &g_volume_parameter;

	if ( NULL == (dp		= find_entry( path )) )
		return ( error );

	if ( !(dp->attributes & 0x10) )
		return ( error );

	make_full_path( full_path, path, vpp->current_dir );

#if BIG_ENDIAN
	vpp->current_directory_cluster	= swap16( dp->cluster );	
#else
	vpp->current_directory_cluster	= dp->cluster;	
#endif

	fs_data_copy( (unsigned char *)(vpp->current_dir), (unsigned char *)full_path, MAX_PATH_STRING_LENGTH );

	return ( no_error );
}


//////////
//////////	"pwd" working directory name
//////////
void fs_pwd( char *pwd_str )
{
	volume_param		*vpp;
	
	vpp				= &g_volume_parameter;

	fs_data_copy( (unsigned char *)pwd_str, (unsigned char *)(vpp->current_dir), MAX_PATH_STRING_LENGTH );
}


//////////
//////////	"ls" get list
//////////
unsigned char fs_ls( char *path, directory_entry *dir, unsigned short entries )
{
	fs_FILE				*fp;
	char				full_path[ MAX_PATH_STRING_LENGTH ];
	unsigned char		*buffer_ptr_base;
	unsigned char		*buffer_ptr;
	volume_param		*vpp;
	unsigned short		i;
	unsigned short		count	= 0;
	
	vpp			= &g_volume_parameter;
	
	make_full_path( full_path, path, vpp->current_dir );

	if ( str_match( full_path, "/", 2 ) )
	{
		//	root directory
		
		entries		= (vpp->root_directory_entries < entries) ? vpp->root_directory_entries : entries;

		buffer_ptr_base	= g_fs_scratch_buffer_ptr;
	
		for ( i = 0; i < entries; i++ )
		{
			if ( !( i % ((vpp->bytes_per_cluster) / sizeof( directory_entry )) ) )
			{
				storage_read_sector( buffer_ptr_base, vpp->start_sector_root_directory + count++, vpp->sector_per_cluster );
				buffer_ptr	= buffer_ptr_base;
			}
			*dir++	= *((directory_entry *)buffer_ptr)++;
		}
	}
	else
	{
		unsigned short	l_read;
		unsigned short	n_read;
	
		if ( NULL == (fp	= fs_fopen( full_path, R | D )) )
		{
			return ( error );
		}	

		l_read	= sizeof( directory_entry ) * entries;
		
		if ( l_read != (n_read		= fs_read( fp, (unsigned char *)dir, l_read )) )
			*((unsigned char *)dir + n_read)	= 0;
		
		fs_fclose( fp );
	}
	
	return ( no_error );
}


//////////
//////////	"fopen" file open
//////////
fs_FILE *fs_fopen( char *path, unsigned char mode )
{
	fs_FILE				*fp;
	directory_entry		*dp;
	volume_param		*vpp;
	unsigned char		index;
	
	vpp				= &g_volume_parameter;

	//	to find open slot to register this open-file

	for ( index = 0; index < MAX_OPEN_FILES; index++ )
		if ( (vpp->opened_FILE_pointer)[ index ] == NULL )
			break;
	
	if ( index == MAX_OPEN_FILES )
	{
//mprintf( WHITE, CONTINUE, "\r\nerror @ fs_fopen : MAX_OPEN_FILES\r\n" );

		g_recent_error	= 2;
		return ( NULL );
	}

	//	allocate fs_FILE structure 

	if ( NULL == ( fp	= (fs_FILE *)malloc( sizeof( fs_FILE ) ) ) )
	{
//mprintf( WHITE, CONTINUE, "\r\nerror @ fs_fopen : malloc1\r\n" );

		g_recent_error	= 3;
		return ( NULL );
	}
	
	//	allocate file buffer (this works as a cache for the disk)
	
	if ( NULL == ( fp->cluster_buffer_pointer	= (unsigned char *)malloc( vpp->bytes_per_cluster ) ) )
	{
//mprintf( WHITE, CONTINUE, "\r\nerror @ fs_fopen : malloc2\r\n" );

		g_recent_error	= 4;
		free( fp );
		return ( NULL );
	}

	make_full_path( fp->path, path, vpp->current_dir );
	separate_directory_and_file( fp->path_to_dir, fp->file_name, fp->path );
	
	if ((mode & 0x01) == FS_WRITE)
	{
		//	if opening for "write", needed to be sure the parent directory exists
	
		if ( NULL == find_entry( fp->path_to_dir ) )
		{
			//	no parent directory
//mprintf( WHITE, CONTINUE, "\r\nerror @ fs_fopen : no parent directory\r\n" );
		
			g_recent_error	= 9;
			free( fp->cluster_buffer_pointer );
			free( fp );
			return ( NULL );
		}
	}
	
	dp	= find_file_entry( path, &(fp->directory_entry_cluster), &(fp->directory_entry_index) );

	if ( (dp->attributes & 0x10) && ((mode & 0x01) != FS_WRITE) )
	{
		if ( (mode & 0x08) != FS_OPN_DIR )
		{
//mprintf( WHITE, CONTINUE, "\r\nerror @ fs_fopen : FS_OPN_DIR\r\n" );
			g_recent_error	= 10;
			free( fp );
			return ( NULL );
		}
	}		
	
	if ( dp )
	{
		fs_data_copy( (unsigned char *)(&(fp->file_info)), (unsigned char *)dp, sizeof( directory_entry ) );

#if BIG_ENDIAN
		(fp->file_info).time		= swap16( (fp->file_info).time );
		(fp->file_info).date		= swap16( (fp->file_info).date );
		(fp->file_info).cluster		= swap16( (fp->file_info).cluster );
		(fp->file_info).size		= swap32( (fp->file_info).size );
#endif
		fp->start_cluster	= (fp->file_info).cluster;
	}
	else
	{
		fp->start_cluster	= 0;
	}

	fp->current_cluster					= 0;
	fp->buffered_cluster				= 0;
	fp->number_of_cluster_processed		= 0;
	fp->current_file_data_position		= 0;
	fp->volume_param_ptr				= vpp;
	fp->index							= index;
	fp->mode							= mode;
	
	(vpp->opened_FILE_pointer)[ index ]	= fp;

	if ((mode & 0x01) == FS_READ)
	{
		//	to open a file to READ

		if  ( !(fp->start_cluster) )
		{
//mprintf( WHITE, CONTINUE, "\r\nerror @ fs_fopen : FS_READ\r\n" );
	
			g_recent_error	= 5;
			free( fp );
			return ( NULL );
		}
	}
	else if ((mode & 0x01) == FS_WRITE)
	{
		//	to open a file to WRITE
	
		if  ( (fp->start_cluster) )
		{
			if ((mode & 0x02) == FS_PROTECT)
			{
				//	for when the over write PROTECTed
//mprintf( WHITE, CONTINUE, "\r\nerror @ fs_fopen : FS_PROTECT\r\n" );
				g_recent_error	= 5;
				free( fp );
				return ( NULL );
			}
			else
			{
				//	for OVERWRITE an existing file		
				fs_rm( fp->path, False );
			}
		}
		
		//	make directry entry for the file
		make_directory_entry_fp( fp );
		
		//	find an open cluster to write file data
		fp->start_cluster	= find_free_cluster( 2 );		//	find open cluster
		write_FAT( fp->start_cluster, 0xFFFF );

		flush_FILE_write_buffer( fp );
	}

	return ( fp );
}


//////////
//////////	"fclose" file close
//////////
void fs_fclose( fs_FILE *fp )
{
	if ( !storage_volume_ready() )
		return;
	
	if ( fp == NULL )
		return;
		
	if ( fp->mode != FS_READ )
	{
		flush_FILE_write_buffer( fp );
		flush_FAT_cache( (volume_param *)(fp->volume_param_ptr) );
		update_directory_entry( fp );
	}

	(((volume_param *)(fp->volume_param_ptr))->opened_FILE_pointer)[ fp->index ]	= NULL;
	
	free( fp->cluster_buffer_pointer );
	free( fp );
}


//////////
//////////	"read" file read
//////////

unsigned short fs_read( fs_FILE *fp, unsigned char *buffer_ptr, unsigned short length )
{
	unsigned short	size_of_pre_read;
	unsigned short	num_of_large_read_clusters;
	unsigned short	size_of_post_read;
	unsigned short	data_offset_in_cluster;
	unsigned short	read_clusters;
	unsigned short	size;
	unsigned short	size_of_total_read	= 0;
	unsigned long	remaining_file_size;

	volume_param	*vpp;

	vpp						= fp->volume_param_ptr;	


	//	calcurate file data size left
	
	if ( (fp->file_info).attributes & 0x10 )
		remaining_file_size		= vpp->bytes_per_cluster;	//	Possible problem : No bigger directory handling
	else
		remaining_file_size		= (fp->file_info).size - fp->current_file_data_position;

	//	limit file reading size to file size.
	//	to prevent over reading.

	length					= ((unsigned long)length < remaining_file_size) ? length : (unsigned short)remaining_file_size;

	if ( length < vpp->bytes_per_cluster )
	{
		size_of_pre_read			= length;
		num_of_large_read_clusters	= 0;
		size_of_post_read			= 0;
	}
	else
	{
		data_offset_in_cluster		= fp->current_file_data_position % vpp->bytes_per_cluster;
		size_of_pre_read			= data_offset_in_cluster ? (vpp->bytes_per_cluster - data_offset_in_cluster) : 0;
		length					   -= size_of_pre_read;
		
		if ( length < vpp->bytes_per_cluster )
		{
			num_of_large_read_clusters	= 0;
			size_of_post_read			= length;
		}
		else
		{
			num_of_large_read_clusters	= length / vpp->bytes_per_cluster;
			size_of_post_read			= length % vpp->bytes_per_cluster;
		}
	}

	if ( size_of_pre_read )
	{
		size	= fs_read_small( fp, buffer_ptr, size_of_pre_read );
		
		buffer_ptr			+= size_of_pre_read;
		size_of_total_read	+= size;
	}
	
	if ( num_of_large_read_clusters )
	{
		size	= fs_read_large( fp, buffer_ptr, num_of_large_read_clusters );
		
		buffer_ptr			+= (num_of_large_read_clusters * vpp->bytes_per_cluster);
		size_of_total_read	+= size;
	}
		
	if ( size_of_post_read )
	{
		size	= fs_read_small( fp, buffer_ptr, size_of_post_read );

		size_of_total_read	+= size;
	}

	return ( size_of_total_read );
}


static unsigned short fs_read_large( fs_FILE *fp, unsigned char *buffer_ptr, unsigned long clusters )
{
	unsigned short	target_cluster_N;
	unsigned short	base_cluster;
	unsigned short	next_cluster;
	unsigned short	read_bytes;
	unsigned short	read_sectors;
	unsigned short	total_read_bytes	= 0;
	unsigned short	continuous_length;
	unsigned short	i;
	volume_param	*vpp;

	vpp				= fp->volume_param_ptr;	

	target_cluster_N		= fp->current_file_data_position / vpp->bytes_per_cluster;
	
	//	cluster buffer control
	//	if the target cluster number does not match to read cluster number,
	//	update the current cluster index.
	
	continuous_length	= 0;
	
	do
	{
		if ( !continuous_length )
		{
			if ( fp->number_of_cluster_processed != (target_cluster_N + 1) )
			{
				if ( fp->current_cluster )
					fp->current_cluster		= read_FAT( fp->current_cluster );
				else
					fp->current_cluster		= fp->start_cluster;
			}
					
			base_cluster		= fp->current_cluster;
		}
		
		continuous_length++;
		clusters--;
		(fp->number_of_cluster_processed)++;
		
		next_cluster		= read_FAT( fp->current_cluster );

		if ( (fp->current_cluster == (next_cluster - 1)) && clusters )
		{
			fp->current_cluster		= next_cluster;
			continue;
		}
		
		read_bytes		= vpp->bytes_per_cluster  * continuous_length;
		read_sectors	= vpp->sector_per_cluster * continuous_length;

		if ( read_sectors != storage_read_sector( buffer_ptr, cluster_to_sector( base_cluster ), read_sectors ) )
		{
			//storage_unmount();
			return ( 0 );
		}

		buffer_ptr			+= read_bytes;
		total_read_bytes	+= read_bytes;
		
		fp->current_file_data_position	+= read_bytes;
		
		continuous_length	= 0;
	}
	while ( clusters );

	return ( total_read_bytes );
}


static unsigned short fs_read_small( fs_FILE *fp, unsigned char *buffer_ptr, unsigned short length )
{
	unsigned long			data_offset_in_cluster;
	unsigned short			target_cluster_N;
	unsigned long			remaining_file_size;
	unsigned short			remaining_size;
	unsigned short			reading_size;
	volume_param			*vpp;

	if ( !length )
		return ( 0 );

	//	get volume parameter pointer to refer volume information.
	//	this function needs information like sector size etc.
	vpp				= fp->volume_param_ptr;	
	
#ifdef ALWAYS_SMALL_READ

	//	calcurate file data size left
	
	if ( (fp->file_info).attributes & 0x10 )
		remaining_file_size		= vpp->bytes_per_cluster;	//	Possible problem : No bigger directory handling
	else
		remaining_file_size		= (fp->file_info).size - fp->current_file_data_position;

	//	limit file reading size to file size.
	//	to prevent over reading.
	length					= ((unsigned long)length < remaining_file_size) ? length : (unsigned short)remaining_file_size;

	//	set "remaining_size". 
	//	the "length" will kept to hold the expected data reading size untill this function end.
	//	it will be used to report read size as return value.
	
#endif // ALWAYS_SMALL_READ

	remaining_size			= length;

	while ( remaining_size )
	{
		//	find target cluster and ofset from data position in the file
	
		data_offset_in_cluster	= fp->current_file_data_position % vpp->bytes_per_cluster;
		target_cluster_N		= fp->current_file_data_position / vpp->bytes_per_cluster;
		
		//	cluster buffer control
		//	if the target cluster number does not match to read cluster number,
		//	update the current cluster index.
		
		if ( fp->number_of_cluster_processed != (target_cluster_N + 1) )
		{

			if ( fp->current_cluster )
			
				//	if it is not first time to load,
				//	get the cluster index from FAT information.

				fp->current_cluster		= read_FAT( fp->current_cluster );
			else
			
				//	if it is first time, set the index which was got in fs_open
				//	"fp->current_cluster" holds the index to the reading source cluster 
			
				fp->current_cluster		= fp->start_cluster;
		}

		//	check the index of reading source and buffered cluster.

		if (fp->current_cluster != fp->buffered_cluster)
		{
			//	if those are different, read from sectors to file buffer
		
			if ( vpp->sector_per_cluster != storage_read_sector( fp->cluster_buffer_pointer, cluster_to_sector( fp->current_cluster ), vpp->sector_per_cluster ) )
			{
				//storage_unmount();
				return ( 0 );
			}
			//	update to show the current buffered cluster information

			fp->buffered_cluster	= fp->current_cluster;
			
			//	count up, the counter to count the cluster has been read.
			
			(fp->number_of_cluster_processed)++;
		}


		//	calculate the readout size. 
		//	The data can be readout from the file data position to cluster end.
	
		reading_size	= (vpp->bytes_per_cluster < (remaining_size + data_offset_in_cluster)) ? (vpp->bytes_per_cluster - data_offset_in_cluster) : remaining_size;
		
		//	read out : 
		//	copy data from file buffer to user buffer for size of reading_size.
		
		//mprintf( LIGHTGRAY, CONTINUE, "remaining_size %d   reading %d   n_cluster %u   offset %ld   c 0x%04X\r\n", remaining_size, reading_size, target_cluster_N, data_offset_in_cluster, fp->current_cluster );
		fs_data_copy( buffer_ptr, (fp->cluster_buffer_pointer) + data_offset_in_cluster, reading_size );
		
		//	update pointers
		//	when the remaining size is not zero, the loop will kepp execution.
		
		fp->current_file_data_position	+= reading_size;
		buffer_ptr						+= reading_size;
		remaining_size					-= reading_size;
	}

	return ( length - remaining_size );
}


//////////
//////////	"write" file write
//////////
//#define ALWAYS_SMALL_WRITE
#ifdef ALWAYS_SMALL_WRITE
#else
unsigned short fs_write( fs_FILE *fp, unsigned char *buffer_ptr, unsigned short length )
{
	unsigned short	size_of_pre_write;
	unsigned short	num_of_large_write_clusters;
	unsigned short	size_of_post_write;
	unsigned short	data_offset_in_cluster;
	unsigned short	write_clusters;
	unsigned short	size;
	unsigned short	size_of_total_write	= 0;
	unsigned long	remaining_file_size;

	volume_param	*vpp;

	vpp						= fp->volume_param_ptr;	

	if ( length < vpp->bytes_per_cluster )
	{
		size_of_pre_write			= length;
		num_of_large_write_clusters	= 0;
		size_of_post_write			= 0;
	}
	else
	{
		data_offset_in_cluster		= fp->current_file_data_position % vpp->bytes_per_cluster;
		size_of_pre_write			= data_offset_in_cluster ? (vpp->bytes_per_cluster - data_offset_in_cluster) : 0;
		length					   -= size_of_pre_write;
		
		if ( length < vpp->bytes_per_cluster )
		{
			num_of_large_write_clusters	= 0;
			size_of_post_write			= length;
		}
		else
		{
			num_of_large_write_clusters	= length / vpp->bytes_per_cluster;
			size_of_post_write			= length % vpp->bytes_per_cluster;
		}
	}

	if ( size_of_pre_write )
	{
		size	= fs_write_small( fp, buffer_ptr, size_of_pre_write );
		
		buffer_ptr			+= size_of_pre_write;
		size_of_total_write	+= size;
	}
	
	if ( num_of_large_write_clusters )
	{
		size	= fs_write_large( fp, buffer_ptr, num_of_large_write_clusters );
		
		buffer_ptr			+= (num_of_large_write_clusters * vpp->bytes_per_cluster);
		size_of_total_write	+= size;
	}
		
	if ( size_of_post_write )
	{
		size	= fs_write_small( fp, buffer_ptr, size_of_post_write );

		size_of_total_write	+= size;
	}

	return ( size_of_total_write );
}


static unsigned short fs_write_large( fs_FILE *fp, unsigned char *buffer_ptr, unsigned long clusters )
{
	unsigned short	target_cluster_N;
	unsigned short	base_cluster;
	unsigned short	next_cluster;
	unsigned short	write_bytes;
	unsigned short	write_sectors;
	unsigned short	total_write_bytes	= 0;
	unsigned short	continuous_length;
	unsigned short	i;
	volume_param	*vpp;

	vpp						= fp->volume_param_ptr;	
	fp->buffered_cluster	= NULL;

	target_cluster_N		= fp->current_file_data_position / vpp->bytes_per_cluster;
	
	//	cluster buffer control
	//	if the target cluster number does not match to write cluster number,
	//	update the current cluster index.
	
	continuous_length	= 0;
	
	do
	{
		if ( !continuous_length )
		{
			if ( fp->number_of_cluster_processed != (target_cluster_N + 1) )
			{
				if ( fp->current_cluster )
				{
					if ( !(next_cluster		= find_free_cluster( fp->current_cluster )) )
						return ( total_write_bytes );
					
					write_FAT( fp->current_cluster, next_cluster );
					write_FAT( next_cluster,        0xFFFF       );		
					fp->current_cluster		= next_cluster;
				}
				else
				{
					fp->current_cluster			= fp->start_cluster;
					(fp->file_info).cluster		= fp->start_cluster;
					
					write_FAT( fp->current_cluster, 0xFFFF       );		
				}
			}
					
			base_cluster		= fp->current_cluster;
		}
		
		continuous_length++;
		clusters--;
		(fp->number_of_cluster_processed)++;
		
		if ( !(next_cluster		= find_free_cluster( fp->current_cluster )) )
			return ( total_write_bytes );

		if ( (fp->current_cluster == (next_cluster - 1)) && clusters )
		{
			write_FAT( fp->current_cluster, next_cluster );
			write_FAT( next_cluster,        0xFFFF       );		
			fp->current_cluster		= next_cluster;
			
			continue;
		}
		
		write_bytes		= vpp->bytes_per_cluster  * continuous_length;
		write_sectors	= vpp->sector_per_cluster * continuous_length;

		if ( write_sectors != storage_write_sector( buffer_ptr, cluster_to_sector( base_cluster ), write_sectors ) )
		{
			//storage_unmount();
			return ( 0 );
		}
		
		write_FAT( fp->current_cluster, 0xFFFF );		

		buffer_ptr			+= write_bytes;
		total_write_bytes	+= write_bytes;
		
		fp->current_file_data_position	+= write_bytes;
		(fp->file_info).size			+= write_bytes;

		continuous_length	= 0;
	}
	while ( clusters );

	return ( total_write_bytes );
}
#endif 	//ALWAYS_SMALL_WRITE



#ifdef ALWAYS_SMALL_WRITE
unsigned short fs_write( fs_FILE *fp, unsigned char *buffer_ptr, unsigned short length )
#else
static unsigned short fs_write_small( fs_FILE *fp, unsigned char *buffer_ptr, unsigned short length )
#endif
{
	unsigned long			data_offset_in_cluster;
	unsigned short			target_cluster_N;
	unsigned short			next_cluster;
	unsigned short			remaining_size;
	unsigned short			writing_size;
	volume_param			*vpp;

	if ( !length )
		return ( 0 );
		
	vpp				= fp->volume_param_ptr;

	remaining_size	= length;

	while ( remaining_size )
	{
		data_offset_in_cluster	= fp->current_file_data_position % vpp->bytes_per_cluster;
		target_cluster_N		= fp->current_file_data_position / vpp->bytes_per_cluster;
				
//		mprintf( LIGHTCYAN, CONTINUE, "remaining_size=%u  data_offset_in_cluster=%lu  target_cluster_N=%u\r\n", remaining_size, data_offset_in_cluster, target_cluster_N );
//		mprintf( LIGHTCYAN, CONTINUE, "fp->start_cluster=%u  fp->number_of_cluster_processed=%u  fp->current_cluster=%u  fp->buffered_cluster=%u\r\n", fp->start_cluster, fp->number_of_cluster_processed, fp->current_cluster, fp->buffered_cluster );

		if ( fp->number_of_cluster_processed != (target_cluster_N + 1) )
		{
			//	if the data on the boundary, new cluster will be allocated

			if ( fp->current_cluster )
			{
				//	if the new cluster required, search the open cluster from current 
				//	cluster.
#if 0			
				if ( !(fp->current_cluster		= find_free_cluster( fp->current_cluster )) )
					return ( length - remaining_size );

#else

				if ( !(next_cluster		= find_free_cluster( fp->current_cluster )) )
					return ( length - remaining_size );
				
				write_FAT( fp->current_cluster, next_cluster );
				write_FAT( next_cluster,        0xFFFF       );		
				fp->current_cluster		= next_cluster;

#endif
			}
			else
			{
				//	the first cluster allocation is done by file open function.
				//	this action just doing the copying the cluster number to current.
				
				fp->current_cluster			= fp->start_cluster;
				(fp->file_info).cluster		= fp->start_cluster;
			}
		}
		

		//	if the curent cluster index and buffered cluster index are different 
		if (fp->current_cluster != fp->buffered_cluster)
		{
			//	if the buffered data already exist
			if ( fp->buffered_cluster )
			{
#define FAST_WRITE
#ifdef FAST_WRITE
				if ( vpp->sector_per_cluster != storage_write_sector( fp->cluster_buffer_pointer, cluster_to_sector( fp->buffered_cluster ), vpp->sector_per_cluster ) )
					return ( 0 );
#else
				flush_FILE_write_buffer( fp );
#endif
				//	set next cluster value to current cluster entry. links the clusters (FATs).
				write_FAT( fp->buffered_cluster, fp->current_cluster );
			}

			write_FAT( fp->current_cluster, 0xFFFF );

			fp->buffered_cluster	= fp->current_cluster;
			(fp->number_of_cluster_processed)++;
		}

		
		writing_size	= (vpp->bytes_per_cluster < (remaining_size + data_offset_in_cluster)) ? (vpp->bytes_per_cluster - data_offset_in_cluster) : remaining_size;

		fs_data_copy( (fp->cluster_buffer_pointer) + data_offset_in_cluster, buffer_ptr, writing_size );

		fp->current_file_data_position	+= writing_size;
		remaining_size					-= writing_size;
		buffer_ptr						+= writing_size;
		(fp->file_info).size			+= writing_size;
	}

	return ( length - remaining_size );
}


static void flush_FILE_write_buffer( fs_FILE *fp )
{
	volume_param			*vpp;

	vpp				= fp->volume_param_ptr;

	if ( fp->buffered_cluster )
		storage_write_sector( fp->cluster_buffer_pointer, cluster_to_sector( fp->buffered_cluster ), vpp->sector_per_cluster );
	
	flush_FAT_cache( vpp );
	update_directory_entry( fp );
}


//////////
//////////	"rm" remove file
//////////
#define		FLAG_LEAVE_CURR_DIR		0x01
#define		FLAG_REMOVE_RECURSIVE	0x02
#define		FLAG_REMOVE_SCAN_ONLY	0x04

unsigned char fs_rm( char *path, unsigned char recursive )
{
	directory_entry		*dp;
	unsigned short		base_cluster;
	unsigned short		index;
	unsigned char		flags	= 0;
	unsigned char		err		= no_error;
	volume_param		*vpp;
	
	vpp				= &g_volume_parameter;

	flags	|= recursive ? FLAG_REMOVE_RECURSIVE : 0;
	
	if ( NULL == (dp	= find_file_entry( path, &base_cluster, &index )) )
		return ( FS_RM_DIRECTORY_ACCESS_ERR );

#if BIG_ENDIAN

	if ( vpp->current_directory_cluster == swap16( dp->cluster ) )
		flags	|= FLAG_LEAVE_CURR_DIR;
#else

	if ( vpp->current_directory_cluster == dp->cluster )
		flags	|= FLAG_LEAVE_CURR_DIR;
		
#endif

	if ( !recursive )
	{
		remove_this_entry( base_cluster, index, flags | FLAG_REMOVE_SCAN_ONLY, &err, 0 );
		
		if ( err )
			return ( err );
	}
	
	remove_this_entry( base_cluster, index, flags, &err, 0 );

	return ( err );
}


//////////
//////////	"mv" move file
//////////
unsigned short fs_mv( char *from, char *to )
{
	fs_FILE				*fp;
	directory_entry		d;
	directory_entry		*dp;
	unsigned short		parent_cluster;
	unsigned short		index;
	unsigned short		i;
//	volume_param		*vpp;
	
//	vpp				= &g_volume_parameter;

	//	get source file entry information

	if ( NULL == (dp	= find_entry( from )) )
		return ( FS_MV_SOURCE_NOT_FOUND );

	fs_data_copy( (unsigned char *)(&d), (unsigned char *)dp, sizeof( directory_entry ) );

	//	make new file entry

	if ( NULL == (fp	= fs_fopen( to, P | W | D )) )
		return ( FS_MV_CANNOT_OPEN_TARGET );
	
	fs_fclose( fp );	//	Just made a directory entry

	if ( NULL == (dp	= find_entry( to )) )
		return ( FS_MV_SOURCE_NOT_FOUND );

	for ( i = 11; i < sizeof( directory_entry ); i++ )
		*((unsigned char *)dp + i)	= *((unsigned char *)(&d) + i);

	write_back_directory_entry();

	//	erase old file entry

	if ( NULL == (dp	= find_entry( from )) )
		return ( FS_MV_SOURCE_NOT_FOUND );

	*((unsigned char *)dp)		= 0xE5;
	
	write_back_directory_entry();

	return ( no_error );
}


//////////
//////////	"mkdir" make directory
//////////
unsigned char fs_mkdir( char *path )
{
	fs_FILE				*fp;
	directory_entry		*dp;
	directory_entry		cur_d		=	{ ".       ", "   ", 0x10, {0,0,0,0,0,0,0,0,0,0}, 0, 0, 0, 0 };
	directory_entry		par_d		=	{ "..      ", "   ", 0x10, {0,0,0,0,0,0,0,0,0,0}, 0, 0, 0, 0 };
	directory_entry		null_d		=	{ {0,0,0,0,0,0,0,0},{0,0,0},0, {0,0,0,0,0,0,0,0,0,0},0,0,0,0 };
	unsigned short		i;
	volume_param		*vpp;
	
	vpp				= &g_volume_parameter;

	//	open directory as writing data into file

	if ( NULL == (fp	= fs_fopen( path, W | D )) )
		return ( error );
	
	//	write directory contents
	
#if BIG_ENDIAN

	cur_d.cluster		= swap16( fp->start_cluster );
	par_d.cluster		= (fp->directory_entry_cluster == 1) ? 0 : fp->directory_entry_cluster;
	par_d.cluster		= swap16( par_d.cluster );

#else

	cur_d.cluster		= fp->start_cluster;
	par_d.cluster		= (fp->directory_entry_cluster == 1) ? 0 : fp->directory_entry_cluster;

#endif

	set_time_for_directory_entry( &cur_d );
	set_time_for_directory_entry( &par_d );
	
	fs_write( fp, (unsigned char *)(&cur_d), sizeof( directory_entry ) );
	fs_write( fp, (unsigned char *)(&par_d), sizeof( directory_entry ) );

	for ( i = 2; i < (vpp->bytes_per_cluster / sizeof( directory_entry )); i++ )
		fs_write( fp, (unsigned char *)(&null_d), sizeof( directory_entry ) );

	//	close written file

	fs_fclose( fp );
	
	
	//	modify file entry

	if ( NULL == (dp	= find_entry( path )) )
		return ( error );

	dp->attributes	= 0x10;
	dp->size		= 0;

	write_back_directory_entry();

	return ( no_error );
}


//////////
//////////	"df" free disk clusters
//////////
void fs_df( unsigned short *available, unsigned short *total, unsigned short *cluster_size )
{
	unsigned short	i;
	
	volume_param	*vpp;

	vpp		= &g_volume_parameter;
	
	*cluster_size	= vpp->bytes_per_cluster;
	*total			= vpp->number_of_clasters;
	*available		= 0;
	
	for ( i = 2; i < (*total) + 2; i++ )
		if ( !read_FAT( i ) )
			(*available)++;
}


#ifdef SUPPORT_DISK_FORMAT

//////////
//////////	"format" format disk
//////////
void fs_format( void )
{
#define		BOOT	0x20
#define		HIDDEN	0x20
#define		CLTSZ	8
#define		ROOTDIRENT	512
#define		RESERVEDSEC	1
#define		N_OF_FAT	2


	MasterBootRecord	*mbrt;
	PartitionEntry		*pet;
	PartitionBootRecord	*pbrt;
	BoisParameterBlock	*bpbt;
	unsigned long		disk_size;
	unsigned long		block_size;
	unsigned long		total_sector;
	unsigned long		available_sector;
	unsigned long		data_sector_size;
	unsigned long		start_sector;

	unsigned short		n_of_cluster;
	unsigned short		FAT_sector_size;
	unsigned char		b[ 512 ];
	unsigned char		blank_str[ 12 ]	= "           ";
	unsigned short		i;


	//	BASIC calc	///////////////////////////////////////////////////
	total_sector	= gp_drive_parameter.last_LBA + 1;
	block_size		= gp_drive_parameter.block_length;
	disk_size		= total_sector * block_size;
	
	available_sector	= total_sector - (BOOT + HIDDEN + RESERVEDSEC + ((sizeof( directory_entry ) * ROOTDIRENT) / block_size));
	n_of_cluster		= (available_sector / CLTSZ) + 2;
	FAT_sector_size		= (((unsigned long)n_of_cluster * 2) / block_size) + 1;	//	(* 2) means the entry has 2 bytes

	data_sector_size	= available_sector - (unsigned long)(FAT_sector_size * N_OF_FAT);
	n_of_cluster		= (data_sector_size / CLTSZ) + 2;
	FAT_sector_size		= (((unsigned long)n_of_cluster * 2) / block_size) + 1;	//	(* 2) means the entry has 2 bytes

#if 0
	mprintf( LIGHTGRAY, CONTINUE, "\r\n" );
	mprintf( LIGHTGRAY, CONTINUE, "total_sector     0x%08lX (%9lu)\r\n", total_sector, total_sector );
	mprintf( LIGHTGRAY, CONTINUE, "block_size       0x%08lX (%9lu)\r\n", block_size, block_size );
	mprintf( LIGHTGRAY, CONTINUE, "disk_size        0x%08lX (%9lu)\r\n", disk_size, disk_size );
	mprintf( LIGHTGRAY, CONTINUE, "available_sector 0x%08lX (%9lu)\r\n", available_sector, available_sector );

	mprintf( LIGHTGRAY, CONTINUE, "n_of_cluster     0x%08X (%9u)\r\n", n_of_cluster, n_of_cluster );
	mprintf( LIGHTGRAY, CONTINUE, "FAT_sector_size  0x%08X (%9u)\r\n", FAT_sector_size, FAT_sector_size );
	mprintf( LIGHTGRAY, CONTINUE, "data_sector_size 0x%08lX (%9lu)\r\n", data_sector_size, data_sector_size );
#endif

	//	MBR		////////////////////////////////////////////////////////

	for ( i = 0; i < 512; i++ )
		*(b + i)	= 0;
	
	mbrt	= (MasterBootRecord *)b;
	mbrt->Signature_Word	= 0xAA55;
	
	pet		= &((mbrt->partition)[ 0 ]);

	pet->DefaultBootPartition		= 0x80;
	pet->StartHead					= 1;
	pet->StartSector				= 1;
	pet->StartCylinder				= 0;				
	pet->PartitionType				= (disk_size >> 25) ? 6 : 4;
	pet->EndHead					= (total_sector >> 14) & 0xFF;
	pet->EndSector					= (total_sector >>  7) & 0x7F;
	pet->EndCylinder				= (total_sector >>  0) & 0x7F;
	pet->StartLBASector				= BOOT;
	pet->NumSectors					= total_sector;

	start_sector	= 0;
	storage_write_sector( b, start_sector, 1 );


	//	PBR		//////////////////////////////////////
	for ( i = 0; i < 512; i++ )
		*(b + i)	= 0;
	
	pbrt	= (PartitionBootRecord *)b;
	bpbt	= &(pbrt->BPB);
	
	(pbrt->JMP_instruction)[ 0 ]	= 0xE9;
	(pbrt->JMP_instruction)[ 1 ]	= 0;
	(pbrt->JMP_instruction)[ 2 ]	= 0;
	fs_data_copy( pbrt->OEM_name_version, blank_str, 8 );
	pbrt->DriveNumber				= 0;
	pbrt->Reserved					= 0;
	pbrt->ExtBootSignature			= 0;
	pbrt->Volume_ID					= 0;
	fs_data_copy( pbrt->VolumeLabel, blank_str, 11 );
	(pbrt->VolumeLabel)[ 0 ]		= 'W';
	(pbrt->VolumeLabel)[ 1 ]		= 'A';
	(pbrt->VolumeLabel)[ 2 ]		= 'S';
	(pbrt->VolumeLabel)[ 3 ]		= 'A';
	(pbrt->VolumeLabel)[ 4 ]		= 'B';
	(pbrt->VolumeLabel)[ 5 ]		= 'I';
	(pbrt->VolumeLabel)[ 6 ]		= '-';
	(pbrt->VolumeLabel)[ 7 ]		= 'D';
	(pbrt->VolumeLabel)[ 8 ]		= 'I';
	(pbrt->VolumeLabel)[ 9 ]		= 'S';
	(pbrt->VolumeLabel)[ 10]		= 'K';
	fs_data_copy( pbrt->FileSysType, blank_str,  8 );
	(pbrt->FileSysType)[ 0 ]		= 'F';
	(pbrt->FileSysType)[ 1 ]		= 'A';
	(pbrt->FileSysType)[ 2 ]		= 'T';
	(pbrt->FileSysType)[ 3 ]		= '1';
	(pbrt->FileSysType)[ 4 ]		= '6';
	*((unsigned short *)(b + 510))	= 0xAA55;

	bpbt->BytesPerSector			= block_size;
	bpbt->SectorsPerClustor			= CLTSZ;
	bpbt->ReservedSectors			= RESERVEDSEC;
	bpbt->NumFATs					= N_OF_FAT;
	bpbt->RootDirEntries			= ROOTDIRENT;
	
	if (disk_size >> 25)
	{
		bpbt->TotalSectors			= 0;
		bpbt->HugeSectors			= data_sector_size;
	}
	else
	{
		bpbt->TotalSectors			= (unsigned short)data_sector_size;
		bpbt->HugeSectors			= 0;
	}
	
	bpbt->MediaIDByte				= 0xF8;
	bpbt->NumFATSectors				= FAT_sector_size;
	bpbt->SectorsPerTrack			= pet->EndSector;
	bpbt->NumHeads					= pet->EndHead;
	bpbt->HiddenSectors				= HIDDEN;
	
	start_sector	= BOOT;
	storage_write_sector( b, start_sector, 1 );
	

	//	FAT1	//////////////////////////////////////
	for ( i = 0; i < 512; i++ )
		*(b + i)	= 0;
	
	start_sector	+= RESERVEDSEC;
	i	= 1;
	while ( i < FAT_sector_size )
	{
		storage_write_sector( b, start_sector + i, 1 );
		i++;
	}
	*(((unsigned short *)b) + 0)	= 0xFFF8;
	*(((unsigned short *)b) + 1)	= 0xFFFF;

	storage_write_sector( b, start_sector, 1 );


	//	FAT2	//////////////////////////////////////
	for ( i = 0; i < 512; i++ )
		*(b + i)	= 0;
	
	start_sector	+= FAT_sector_size;
	i	= 1;
	while ( i < FAT_sector_size )
	{
		storage_write_sector( b, start_sector + i, 1 );
		i++;
	}
	*(((unsigned short *)b) + 0)	= 0xFFF8;
	*(((unsigned short *)b) + 1)	= 0xFFFF;

	storage_write_sector( b, start_sector, 1 );


	//	ROOT DIR	//////////////////////////////////
	for ( i = 0; i < 512; i++ )
		*(b + i)	= 0;
	
	start_sector	+= FAT_sector_size;

	for ( i = 0; i < (ROOTDIRENT * sizeof( directory_entry )) / block_size; i++ )
		storage_write_sector( b, start_sector++, 1 );
}

#endif

unsigned long fs_cluster_to_sector( unsigned short c )
{
	return ( cluster_to_sector( c ) );
}


unsigned short fs_read_FAT( unsigned short current_cluster )
{
	return ( read_FAT( current_cluster ) );
}




/*
==============================================================================
==============================================================================
====	Public functions for file system access
====		Functions for DRIVER side
==============================================================================
==============================================================================
*/

volume_param *fs_volume_open( void )
{
	volume_param	*vpp;
	char			fs_name[ 9 ];
	unsigned char	*buffer_ptr;
	unsigned char	partition;		//	##	Version 1.2c
	unsigned char	fs_type;
	unsigned long	start_lba;
	unsigned char	i;

	//	Make volume instance
	//	In this version, it is allocated as static

	vpp				= &g_volume_parameter;


	//	Allocate a temporary buffer for sector size
	//	At this moment, the cluster size is unknown. So this temporary buffer will be freed 
	//	after getting cluster size information from boot block in the volume.

	if ( NULL == (buffer_ptr	= (unsigned char *)malloc( gp_drive_parameter.block_length )) )
	{
		mprintf( LIGHTGRAY, CONTINUE, "error : File system init (malloc) failed\r\n" );
		return ( NULL );
	}
	

	//	Read sector LBA=0 (MasterBootRecord : MBR) to get partition information (Partition entries).
	//	This version only support single partition.
	//	MBR : PC card standard section 3.2 and 3.2.4 (include "Partition entry" information)
	//	Also, see http://www.microsoft.com/hwdev/download/hardware/FATGEN103.doc
		
	storage_read_sector( buffer_ptr, 0, 1 );
	
	
	//	##	Version 1.2c (from here)
	//
	//		When the storage volume does not have partition table, it should be 
	//		handled as a volume without partition table. 
	//		Some storage which has been formatted by Windows may not be having the 
	//		table. 
	//		
	//		Following is the comment from user. 
         
	//		# Now, I figured out FAT16 filesystem bug that I mentioned at previous 
	//		# mail.
	//		# 
	//		# The problem is, when formatting disk in Windows with FAT16, 
	//		# sometimes disk has no Master Boot Record, but only have Partition 
	//		# Boot Record. i.e. Disk has only one partition like floppy disk. 
	//		# In this case, skipping MBR and direct reading PBR is needed.

	partition	= ((((MasterBootRecord *)buffer_ptr)->partition)[ 0 ]).DefaultBootPartition;
	
	if ( (partition & ~0x80) == 0x00 )	//	partition table exists
	{
		fs_type		= ((((MasterBootRecord *)buffer_ptr)->partition)[ 0 ]).PartitionType;
		start_lba	= ((((MasterBootRecord *)buffer_ptr)->partition)[ 0 ]).StartLBASector;
		
#if BIG_ENDIAN
		start_lba	= swap32( start_lba );
#endif
	
		//	The parttition information gives you the sector (LBA) number which is the starting point 
		//	of valid partition data. Read the first sector of the partition.
		//	This sector has "PartitionBootRecord (PBR)".
		//	PBR : PC card standard section 4.1.7.1 (include "Partition entry" information)
		
		storage_read_sector( buffer_ptr, start_lba, 1 );
	}	
	//	##	Version 1.2c (to here)

	//	Setting volume information. It stores size of units, and address of boundaries.
	//
	//	The volume will have logical structure like...
	/*
	
			+-----------------------------+ -- 0x00000000
			|    MasterBootRecord         |
			+ - - - - - - - - - - - - - - +
			|                             |
			|                             |
			+-----------------------------+ -- start of partition : (MasterBootRecord.partition[ 0 ]).StartLBASector
			|    PartitionBootRecord      |
			+ - - - - - - - - - - - - - - +
			|                             |
			|                             |
			+-----------------------------+ -- start of FAT1 = start of partition + (PartitionBootRecord.BIOS_parameter_block).ReservedSectors
			|    FAT1                     |
			|                             |
			|                             |
			+-----------------------------+ -- start of FAT2 = start of FAT1 + (PartitionBootRecord.BIOS_parameter_block).NumFATSectors
			|    FAT2                     |
			|                             |
			|                             |
			+-----------------------------+ -- start of RootDirectory = start of FAT2 + (PartitionBootRecord.BIOS_parameter_block).NumFATSectors
			|    RootDirectory            |
			|                             |
			|                             |
			|                             |
			+-----------------------------+ -- start of File/Directory data = start of RootDirectory + (PartitionBootRecord.BIOS_parameter_block).RootDirEntries * sizeof( directory_entry )
			|    File/Directory data      |
			|                             |
			|                             |
			|                             |
			|                             |
			+ - - - - - - - - - - - - - - +	 <---
			|                             |    |
			+    -    -    -    -    -    +    |  size of claster = (PartitionBootRecord.BIOS_parameter_block).SectorsPerClustor
			|                             |    |
			+ - - - - - - - - - - - - - - +  <---
			|                             |    |  size of sector =  (PartitionBootRecord.BIOS_parameter_block).BytesPerSector
			+    -    -    -    -    -    +  <---                        ... This can be get by SCSI command "READ_CAPACITY" command also.
			|                             |
			+ - - - - - - - - - - - - - - +
			|                             |
	
	*/

#if BIG_ENDIAN
	fs_data_copy( fs_name, buffer_ptr + 0x36, 8 );
#else
	fs_data_copy( fs_name, ((PartitionBootRecord *)buffer_ptr)->FileSysType, 8 );
#endif

	*(fs_name + 8)	= 0;

	vpp->file_system		= FILE_SYSTEM_UNKNOWN;

	if ( str_match( fs_name, "FAT16   ", 8 ) )
		vpp->file_system		= FILE_SYSTEM_FAT16;

	if ( str_match( fs_name, "FAT12   ", 8 ) )
		vpp->file_system		= FILE_SYSTEM_FAT12;
	
	if ( vpp->file_system == FILE_SYSTEM_UNKNOWN )
	{
		mprintf( LIGHTGRAY, CONTINUE, "error : Unsupported file system [0x%02X]\r\n", fs_type );
		return ( NULL );
	}
	else
	{

#if 0
		mprintf( LIGHTRED, CONTINUE, "FILE SYSTEM NAME=\"%s\"\r\n", fs_name );
#endif
	}

#if BIG_ENDIAN

	vpp->bytes_per_sector				= *((unsigned short *)(buffer_ptr + 0xB + 0x00));	//	BytesPerSector;
	vpp->bytes_per_sector				= swap16( vpp->bytes_per_sector );
	
	vpp->sector_per_cluster				= *((unsigned char  *)(buffer_ptr + 0xB + 0x02));	//	SectorsPerClustor;

	if ( !vpp->sector_per_cluster )
		return ( NULL );

	vpp->bytes_per_cluster				= vpp->bytes_per_sector * vpp->sector_per_cluster;
	
	vpp->reserved_sectors				= *((unsigned short *)(buffer_ptr + 0xB + 0x03));	//	ReservedSectors;
	vpp->reserved_sectors				= swap16( vpp->reserved_sectors );
	
	vpp->number_of_FATs					= *((unsigned char  *)(buffer_ptr + 0xB + 0x05));	//	NumFATs;

	vpp->root_directory_entries			= *((unsigned short *)(buffer_ptr + 0xB + 0x06));	//	RootDirEntries;
	vpp->root_directory_entries			= swap16( vpp->root_directory_entries );
	
	vpp->total_sectors					= *((unsigned short *)(buffer_ptr + 0xB + 0x08));	//	TotalSectors;
	vpp->total_sectors					= swap16( vpp->total_sectors );

	vpp->number_of_FAT_sectors			= *((unsigned short *)(buffer_ptr + 0xB + 0x0B));	//	NumFATSectors;
	vpp->number_of_FAT_sectors			= swap16( vpp->number_of_FAT_sectors );
	
	if ( !(vpp->total_sectors) )
	{
		vpp->total_sectors					= *((unsigned long *)(buffer_ptr + 0xB + 0x15));	//	HugeSectors;
		vpp->total_sectors					= swap32( vpp->total_sectors );
	}


#else

	vpp->bytes_per_sector				= (((PartitionBootRecord *)buffer_ptr)->BPB).BytesPerSector;
	vpp->sector_per_cluster				= (((PartitionBootRecord *)buffer_ptr)->BPB).SectorsPerClustor;

	if ( !vpp->sector_per_cluster )
		return ( NULL );

	vpp->bytes_per_cluster				= vpp->bytes_per_sector * vpp->sector_per_cluster;
	
	vpp->reserved_sectors				= (((PartitionBootRecord *)buffer_ptr)->BPB).ReservedSectors;
	
	vpp->number_of_FATs					= (((PartitionBootRecord *)buffer_ptr)->BPB).NumFATs;
	vpp->root_directory_entries			= (((PartitionBootRecord *)buffer_ptr)->BPB).RootDirEntries;
	vpp->number_of_FAT_sectors			= (((PartitionBootRecord *)buffer_ptr)->BPB).NumFATSectors;
	vpp->total_sectors					= (((PartitionBootRecord *)buffer_ptr)->BPB).TotalSectors;
	
	if ( !(vpp->total_sectors) )
		vpp->total_sectors					= (((PartitionBootRecord *)buffer_ptr)->BPB).HugeSectors;

#endif
	
	free( buffer_ptr );

	//	The temporary buffer freed.
	//	The sector size data buffering will not be used further.
	
	
	vpp->start_sector_FAT1				= start_lba + vpp->reserved_sectors;
	vpp->start_sector_FAT2				= vpp->start_sector_FAT1 + vpp->number_of_FAT_sectors;
	vpp->start_sector_root_directory	= vpp->start_sector_FAT2 + vpp->number_of_FAT_sectors;
	vpp->start_sector_claster_region	= vpp->start_sector_root_directory + ((sizeof( directory_entry ) * vpp->root_directory_entries) / vpp->bytes_per_sector);
	vpp->number_of_clasters				= (vpp->total_sectors - vpp->start_sector_claster_region) / vpp->sector_per_cluster;
	
	if ( (sizeof( directory_entry ) * vpp->root_directory_entries) % vpp->bytes_per_sector )
		vpp->start_sector_claster_region	+= 1; 
	
	
	//	Allocate FAT cache

	if ( vpp->file_system == FILE_SYSTEM_FAT16 )
	{
		//	For FAT16, the FAT would be too big to load whole size.
		//	At this stage, the driver allocates the memory only.
		//	The FAT is loaded and stored in sector base when it is used.

		if ( NULL == (vpp->FAT_cache	= (unsigned short *)malloc( vpp->bytes_per_sector )) )
		{
			mprintf( LIGHTGRAY, CONTINUE, "error : File system init (malloc 3rd) failed\r\n" );
			return ( NULL );
		}
	}
	else if ( vpp->file_system == FILE_SYSTEM_FAT12 )
	{
		if ( NULL == (vpp->FAT_cache	= (unsigned short *)malloc( vpp->bytes_per_sector * vpp->number_of_FAT_sectors )) )
		{
			mprintf( LIGHTGRAY, CONTINUE, "error : File system init (malloc 3rd) failed\r\n" );
			return ( NULL );
		}

		storage_read_sector( (unsigned char *)(vpp->FAT_cache), vpp->start_sector_FAT1, vpp->number_of_FAT_sectors );
	}
	
	vpp->FAT_cached_sector	= FAT_CACHE_NOT_VALID;
	vpp->FAT_cache_dirty	= False;


	//	Allocate directory cache

	if ( NULL == (vpp->dir_cache	= (unsigned char *)malloc( vpp->bytes_per_cluster )) )
	{
		mprintf( LIGHTGRAY, CONTINUE, "error : File system init (malloc 4th) failed\r\n" );
		return ( NULL );
	}
	vpp->dir_cached_sector	= DIR_CACHE_NOT_VALID;


	//	Clear file open entries

	for ( i = 0; i < MAX_OPEN_FILES; i++ )
		(vpp->opened_FILE_pointer)[ i ]			= NULL;
		
		
	//	Set default current directory to root
	
	(vpp->current_dir)[ 0 ]				= '/';
	(vpp->current_dir)[ 1 ]				= 0;
	vpp->current_directory_cluster		= 0;

	//	Scratch buffer allocate
	
	if ( g_num_of_fs_client )
	{
		if ( g_fs_scratch_buffer_size < vpp->bytes_per_cluster )
		{
			free( g_fs_scratch_buffer_ptr );
		
			if ( NULL == ( g_fs_scratch_buffer_ptr	= (unsigned char *)malloc( vpp->bytes_per_cluster ) ) )
			{
				mprintf( LIGHTGRAY, CONTINUE, "error : File system init (malloc for FS scratch) failed\r\n" );
				return ( NULL );
			}
			
			g_fs_scratch_buffer_size	= vpp->bytes_per_cluster;
		}
	}
	
	if ( !g_num_of_fs_client )
	{
		if ( NULL == ( g_fs_scratch_buffer_ptr	= (unsigned char *)malloc( vpp->bytes_per_cluster ) ) )
		{
			mprintf( LIGHTGRAY, CONTINUE, "error : File system init (malloc for FS scratch) failed\r\n" );
			return ( NULL );
		}
	}
		
	g_num_of_fs_client++;

	return ( vpp );
}


unsigned char fs_volume_close( volume_param *vpp )
{
	unsigned char	i;

	if ( vpp == NULL )
		return ( no_error );
		
	for ( i = 0; i < MAX_OPEN_FILES; i++ )
		if ( (vpp->opened_FILE_pointer)[ i ] != NULL )
			fs_fclose( (vpp->opened_FILE_pointer)[ i ] );
	
	free( vpp->dir_cache );
	free( vpp->FAT_cache );

	g_num_of_fs_client--;

	if ( !g_num_of_fs_client )
		free( g_fs_scratch_buffer_ptr );
	
	//	free( vpp );
	
	return ( no_error );
}



/*
==============================================================================
==============================================================================
====	Public function for file system access
====		utility
==============================================================================
==============================================================================
*/


unsigned short fs_data_copy( unsigned char *trg, unsigned char *src, unsigned short n )
{
	unsigned short	i;

	for ( i = 0; i < n; i++ )
		*trg++		= *src++;
	
	return ( i );
}



/*
==============================================================================
==============================================================================
====	Private functions for file system access
====		accessing media
==============================================================================
==============================================================================
*/


static unsigned char make_directory_entry_fp( fs_FILE *fp )
{
	return (make_directory_entry( fp->path, fp->directory_entry_cluster, &(fp->directory_entry_index), &(fp->file_info), NO_MAKE_NEW_DIR_ENT ) );	
}


static unsigned char make_directory_entry( char *full_path, unsigned short parent_cluster, unsigned short *index_ptr, directory_entry *dir_ptr, unsigned char make_dir )
{
	directory_entry		*dp;
	char                name[ 12 ];
	unsigned char		i;
	volume_param		*vpp;
	
	if ( find_entry( full_path ) )
		return ( ERROR_MAKE_DIR__ENTRY_EXIST );

	format_file_name_directory_entry( name, full_path + 1 + last_char_position( full_path, '/' ) );

	if ( NULL == (dp	= find_free_directory_entry( index_ptr, parent_cluster, name )) )
		return ( error );

	//	initialize all field

	fs_data_copy( (unsigned char *)dp, name, 11 );
	
	dp->attributes		= make_dir ? 0x10 : 0x20;
	
	for ( i = 0; i < 10; i++ )
		(dp->reserved)[ i ]		= 0;
		
	set_time_for_directory_entry( dp );

	dp->cluster			= 0x0000;
	dp->size			= 0;
	
	//	copy to file parameter structure
	
	if ( dir_ptr )
		fs_data_copy( (unsigned char *)dir_ptr, (unsigned char *)dp, sizeof( directory_entry ) );

	write_back_directory_entry();

	return ( no_error );
}


#if BIG_ENDIAN


static unsigned char update_directory_entry( fs_FILE *fp )
{
	directory_entry		*dp;
	directory_entry		tmp_d;
	
	fs_data_copy( (unsigned char *)(&tmp_d), (unsigned char *)(&(fp->file_info)), sizeof( directory_entry ) );

	tmp_d.time		= swap16( tmp_d.time );
	tmp_d.date		= swap16( tmp_d.date );
	tmp_d.cluster	= swap16( tmp_d.cluster );
	tmp_d.size		= swap32( tmp_d.size );
	
	if ( NULL == (dp	= get_directory_entry( fp->directory_entry_cluster, fp->directory_entry_index, NO_MAKE_NEW_DIR_ENT )) )
		return ( error );

	fs_data_copy( (unsigned char *)dp, (unsigned char *)(&tmp_d), sizeof( directory_entry ) );
	
	write_back_directory_entry();

	return ( no_error );
}


#else


static unsigned char update_directory_entry( fs_FILE *fp )
{
	directory_entry		*dp;

	if ( NULL == (dp	= get_directory_entry( fp->directory_entry_cluster, fp->directory_entry_index, NO_MAKE_NEW_DIR_ENT )) )
		return ( error );

	fs_data_copy( (unsigned char *)dp, (unsigned char *)(&(fp->file_info)), sizeof( directory_entry ) );
	
	write_back_directory_entry();

	return ( no_error );
}


#endif


static unsigned char remove_this_entry( unsigned short base_cluster, unsigned short index, unsigned char flags, unsigned char *ep, unsigned short depth )
{
	directory_entry		*dp;
	unsigned short		sub_base;
	unsigned char		c;
	unsigned short		i;

	if ( NULL == (dp	= get_directory_entry( base_cluster, index, NO_MAKE_NEW_DIR_ENT )) )
		return ( True );

	c	= *((unsigned char *)dp);
	
	if ( c == 0 )
		return ( True );
		
	if ( dp->attributes == 0x0F )
		c	= 0xE5;
	
	if ( depth && !(flags & (FLAG_REMOVE_RECURSIVE | FLAG_LEAVE_CURR_DIR)) && (c != 0xE5) )
	{
		*ep		= FS_RM_SUB_DIRECTORY_IS_NOTEMPTY;
		return ( True );
	}
	
	if ( dp->attributes & 0x10 )
	{

#if BIG_ENDIAN
		sub_base	= swap16( dp->cluster );
#else
		sub_base	= dp->cluster;
#endif

		if ( !depth || (flags & FLAG_REMOVE_RECURSIVE) || (c == 0xE5) )
		{
			i	= (sub_base == ROOT_DIRECTORY_CLUSTER) ? 0 : 2;
			
			while ( !remove_this_entry( sub_base, i++, flags, ep, depth + 1 ) )
				;
		}
		else
		{
			*ep		= FS_RM_SUB_DIRECTORY_EXIST;
			return ( True );
		}
	}
	
	dp	= get_directory_entry( base_cluster, index, NO_MAKE_NEW_DIR_ENT );
	
	if ( *ep )
		return ( *ep );

	if ( !(flags & FLAG_REMOVE_SCAN_ONLY) && !(!depth && (flags & FLAG_LEAVE_CURR_DIR)) )
	{
		*((unsigned char *)dp)	= 0xE5;
		write_back_directory_entry();

#if BIG_ENDIAN
		clear_FAT_chain( swap16( dp->cluster ) );
#else
		clear_FAT_chain( dp->cluster );
#endif
	}

	return ( no_error );
}









//
//	FAT operation
//
static unsigned short count_free_cluster( void )
{
	unsigned short	count	= 0;
	unsigned short	i;
	
	volume_param	*vpp;

	vpp		= &g_volume_parameter;

	for ( i = 0; i < vpp->number_of_clasters; i++ )
		if ( !read_FAT( i + 2 ) )
			count++;
	
	return ( count );
}


static unsigned short find_free_cluster( unsigned short start_cluster )
{
	unsigned short	i;
	
	volume_param	*vpp;

	vpp		= &g_volume_parameter;

	for ( i = start_cluster; i < (vpp->number_of_clasters + 2); i++ )
		if ( !read_FAT( i ) )
			return ( i );
	
	return ( 0 );
}


static unsigned char clear_FAT_chain( unsigned short cluster )
{
	unsigned short	count;
	unsigned short	i;
	volume_param	*vpp;

	vpp		= &g_volume_parameter;

	count	= 0;
	while ( valid_cluster( cluster, vpp ) )
	{
		cluster		= write_FAT( cluster, 0x0000 );

		if ( count++ == 0xFFFF )
			break;
	}

	flush_FAT_cache( vpp );
	
	if ( (i == 0xFFFF) || (cluster == 0x0000) )
		return ( error );

	return ( no_error );
}


static unsigned char valid_cluster( unsigned short cluster, volume_param *vpp )
{
	if ( (cluster - 2) < vpp->number_of_clasters )
		return ( True );
	
	return ( False );
}


static unsigned short read_FAT( unsigned short current_cluster )
{
	volume_param			*vpp;

	vpp				= &g_volume_parameter;

	if ( vpp->file_system == FILE_SYSTEM_FAT16 )
		return ( access_FAT16( FAT_READ, current_cluster, NULL ) );
		
	else if ( vpp->file_system == FILE_SYSTEM_FAT12 )
		return ( access_FAT12( FAT_READ, current_cluster, NULL ) );
	
	return ( 0 );
}


static unsigned short write_FAT( unsigned short cluster, unsigned short value )
{
	volume_param			*vpp;

	vpp				= &g_volume_parameter;

	if ( vpp->file_system == FILE_SYSTEM_FAT16 )
		return ( access_FAT16( FAT_WRITE, cluster, value ) );
		
	else if ( vpp->file_system == FILE_SYSTEM_FAT12 )
		return ( access_FAT12( FAT_WRITE, cluster, value ) );

	return ( 0 );
}


static unsigned short flush_FAT_cache( volume_param *vpp )
{
	if ( vpp->file_system == FILE_SYSTEM_FAT16 )
		return ( flush_FAT16_cache( vpp ) );
		
	else if ( vpp->file_system == FILE_SYSTEM_FAT12 )
		return ( flush_FAT12_cache( vpp ) );

	return ( 0 );
}


static unsigned short flush_FAT16_cache( volume_param *vpp )
{
	if ( vpp->FAT_cached_sector == FAT_CACHE_NOT_VALID )
		return ( 0 );
		
	storage_write_sector( (unsigned char *)(vpp->FAT_cache), vpp->start_sector_FAT1 + vpp->FAT_cached_sector, 1 );
	storage_write_sector( (unsigned char *)(vpp->FAT_cache), vpp->start_sector_FAT2 + vpp->FAT_cached_sector, 1 );
	
	vpp->FAT_cached_sector		= FAT_CACHE_NOT_VALID;
	vpp->FAT_cache_dirty		= False;

	return ( 0 );
}


static unsigned short flush_FAT12_cache( volume_param *vpp )
{
	if ( !(vpp->FAT_cache_dirty) )
		return ( 0 );

	storage_write_sector( (unsigned char *)(vpp->FAT_cache), vpp->start_sector_FAT1, vpp->number_of_FAT_sectors );
	storage_write_sector( (unsigned char *)(vpp->FAT_cache), vpp->start_sector_FAT2, vpp->number_of_FAT_sectors );
	
	vpp->FAT_cache_dirty		= False;

	return ( 0 );
}


static unsigned short access_FAT16( unsigned char mode, unsigned short cluster, unsigned short value )
{
	unsigned long			target_sector;
	unsigned long			entry_offset;
	unsigned short			previous_value;
	volume_param			*vpp;

	vpp				= &g_volume_parameter;

	target_sector	= cluster / (vpp->bytes_per_sector >> 1);
	entry_offset	= cluster % (vpp->bytes_per_sector >> 1);
	
	if ( vpp->FAT_cached_sector != target_sector )
	{
		if ( vpp->FAT_cache_dirty )
			flush_FAT_cache( vpp );

		storage_read_sector( (unsigned char *)(vpp->FAT_cache), vpp->start_sector_FAT1 + target_sector, 1 );
		vpp->FAT_cached_sector		= target_sector;
	}
	
	if ( mode == FAT_WRITE )
	{
#if BIG_ENDIAN

		previous_value		= *(vpp->FAT_cache + entry_offset);
		*(vpp->FAT_cache + entry_offset)	= swap16( value );
		vpp->FAT_cache_dirty				= True;
		return ( swap16( previous_value ) );
		
#else

		previous_value		= *(vpp->FAT_cache + entry_offset);
		*(vpp->FAT_cache + entry_offset)	= value;
		vpp->FAT_cache_dirty				= True;
		return ( previous_value );
		
#endif
	}		
	
	//else if ( mode == FAT_READ  )
#if BIG_ENDIAN
		return ( swap16( *(vpp->FAT_cache + entry_offset) ) );
#else
		return ( *(vpp->FAT_cache + entry_offset) );
#endif
}


static unsigned short access_FAT12( unsigned char mode, unsigned short cluster, unsigned short value )
{
	unsigned long			*target;
	unsigned long			offset;
	unsigned char			odd;
	unsigned long			tmp;
	unsigned short			target_cluster;
	volume_param			*vpp;

	vpp				 = &g_volume_parameter;

	odd				 = cluster & 0x1;
	cluster			&= 0xFFF;
	offset			 = (cluster >> 1) * 3;

	target	= (unsigned long *)(((unsigned char *)vpp->FAT_cache) + offset);

#if BIG_ENDIAN
	tmp		= swap32( *target );
#else
	tmp		= *target;
#endif
	
	target_cluster	 = odd ? ((tmp >> 12) & 0xFFF) : (tmp & 0xFFF);
	
	target_cluster	|= (target_cluster & 0x0800) ? 0xF000 : 0x0000;

	if ( mode == FAT_WRITE )
	{
		value			&= 0xFFF;
		
		if ( odd )
		{
			tmp		&= 0xFF000FFF;
			tmp		|= (unsigned long)value << 12;
		}
		else
		{
			tmp		&= 0xFFFFF000;
			tmp		|= (unsigned long)value;
		}

#if BIG_ENDIAN
		*target					= swap32( tmp );
#else
		*target					= tmp;
#endif

		vpp->FAT_cache_dirty	= True;
	}
	
	return ( target_cluster );
}


//
//	Sector operation
//
static unsigned long cluster_to_sector( unsigned short cluster )
{
	volume_param			*vpp;

	vpp				= &g_volume_parameter;

	if ( cluster == ROOT_DIRECTORY_CLUSTER )
		return ( vpp->start_sector_root_directory );
	else if ( valid_cluster( cluster, vpp ) )
		return ( vpp->start_sector_claster_region + ((((unsigned long)cluster) - 2) * (unsigned long)(vpp->sector_per_cluster)) );
	else
		return ( 0xFFFFFFFF );
}

//
//	Cluster operation
//
static directory_entry *find_entry( char *path )
{
	unsigned short dummy_c;
	unsigned short dummy_i;
	
	return ( find_file_entry( path, &dummy_c, &dummy_i ) );
}


static directory_entry *find_file_entry( char *path, unsigned short *dir_cluster_ptr, unsigned short *index_ptr )
{
	char				full_path[ MAX_PATH_STRING_LENGTH ];
	char				*fpp;
	char				file_entry_string[ 12 ];
	directory_entry		*dp;
	volume_param		*vpp;

	vpp					= &g_volume_parameter;

	dp					= &g_root_dumy;
	
#if BIG_ENDIAN
	*dir_cluster_ptr	= swap16( dp->cluster );
#else
	*dir_cluster_ptr	= dp->cluster;
#endif

	fpp		= full_path;
	
	make_full_path( fpp, path, vpp->current_dir );

	while ( 1 )
	{
		if ( *fpp == '/' )
			fpp++;

		if ( !(*fpp) )
			break;
			
		format_file_name_directory_entry( file_entry_string, fpp );	//	This formats the name first directory name in the path

#if BIG_ENDIAN

		*dir_cluster_ptr	= swap16( dp->cluster );

		if ( NULL == (dp	= find_directory_index( index_ptr, swap16( dp->cluster ), file_entry_string )) )
			break;
#else
		*dir_cluster_ptr	= dp->cluster;

		if ( NULL == (dp	= find_directory_index( index_ptr, dp->cluster, file_entry_string )) )
			break;
#endif


		fpp			+= first_char_position( fpp, '/' );
	}
	
	return ( dp );
}


static directory_entry *find_directory_index( unsigned short *found_index, unsigned short base_cluster, char *name )
{
	directory_entry		*dp;
	
	*found_index	= 0;

	while ( 1 )
	{
		dp		= get_directory_entry( base_cluster, *found_index, NO_MAKE_NEW_DIR_ENT );
		
		if ( NULL == dp )
			return ( NULL );

		if ( *((unsigned char *)dp) == 0 )
			return ( NULL );
			
		if ( str_match( (char *)dp, name, 11 ) )
			return ( dp );
		
		(*found_index)++;
	}
}


static directory_entry *find_free_directory_entry( unsigned short *found_index, unsigned short base_cluster, char *name )
{
	directory_entry		*dp;
	
	*found_index	= 0;
	
	while ( 1 )
	{
		dp		= get_directory_entry( base_cluster, *found_index, OK_MAKE_NEW_DIR_ENT );
		
		if ( NULL == dp )
			return ( NULL );

		if ( *((unsigned char *)dp) == 0 )
			return ( dp );
			
		if ( str_match( (char *)dp, name, 11 ) )
			return ( NULL );
		
		if ( *((unsigned char *)dp) == 0xE5 )
			return ( dp );
			
		(*found_index)++;
	}
}


static directory_entry *get_directory_entry( unsigned short base_cluster, unsigned short index, unsigned char new_entry_ok )
{
	unsigned short			cluster_offset;
	unsigned short			entry_offset;
	unsigned short			target_cluster;
	unsigned short			new_cluster;
	unsigned long			read_start_sector		= 0;
	unsigned char			new_sector_allocated	= False;
	unsigned short			i;
	volume_param			*vpp;

	vpp					= &g_volume_parameter;
	
	entry_offset		= index % (vpp->bytes_per_cluster / sizeof(directory_entry));
	cluster_offset		= index / (vpp->bytes_per_cluster / sizeof(directory_entry));

	if ( base_cluster == ROOT_DIRECTORY_CLUSTER )
	{
		read_start_sector	= vpp->start_sector_root_directory + vpp->sector_per_cluster * cluster_offset;
	}
	else
	{
		target_cluster	= base_cluster;
		
		for ( i = 0; i < cluster_offset; i++ )
		{
			new_cluster		= read_FAT( target_cluster );
			
			if ( new_cluster == 0xFFFF )
			{
				if ( new_entry_ok )
				{
					new_cluster		= find_free_cluster( target_cluster );
					
					if ( !valid_cluster( new_cluster, vpp ) )
						return ( NULL );
						
					write_FAT( target_cluster, new_cluster );
					write_FAT( new_cluster, 0xFFFF );
					new_sector_allocated	= True;
				}
				else
				{
					return ( NULL );
				}
			}
			
			target_cluster	= new_cluster;
		}
	
		read_start_sector	= cluster_to_sector( target_cluster );
	}
	
	if ( read_start_sector != vpp->dir_cached_sector )
	{
		if ( new_sector_allocated )
			for ( i = 0; i < vpp->bytes_per_cluster; i++ )
				*(vpp->dir_cache + i)	= 0;
		else
			storage_read_sector( vpp->dir_cache, read_start_sector, vpp->sector_per_cluster );
		
		vpp->dir_cached_sector	= read_start_sector;
	}
	return ( (directory_entry *)(vpp->dir_cache + (entry_offset * sizeof(directory_entry))) );
}


static void write_back_directory_entry( void )
{
	volume_param		*vpp;

	vpp					= &g_volume_parameter;	
	
	storage_write_sector( vpp->dir_cache, vpp->dir_cached_sector, vpp->sector_per_cluster );
}




/*
==============================================================================
==============================================================================
====	Private functions for file system access
====		utility functions
==============================================================================
==============================================================================
*/


#include	<time.h>

static void set_time_for_directory_entry( directory_entry *d )
{
	time_t		ti;
	struct tm	*t;
	
	ti	= time( NULL );
	t	= localtime( &ti );
	
	d->time		= ((t->tm_hour << 11) & 0xF800) | ((t->tm_min << 5) & 0x07E0) | ((t->tm_sec >> 1) & 0x001F);
	d->date		= (((t->tm_year - 80) << 9) & 0xFE00) | (((t->tm_mon + 1) << 5) & 0x01E0) | (t->tm_mday & 0x001F);
}


static unsigned char format_file_name_directory_entry( char *target_base, char *source )
{
	char			*target;
	unsigned char	count	= 0;
	char			c;
	
	target	= target_base;
	
	while ( target < (target_base + 11) )
		*target++	= ' ';
		
	target	= target_base;
	
//	while ( c	= *source++ )
	while ( 0 != (c	= *source++) )
	{
		if (count++ == 12)
			break;
			
		if (c == '/')
			break;
			
		if (c == '.')
		{
			target	= target_base + 8;	
			continue;
		}
		
		if ( ('a' <= c) && (c <= 'z') )
			c	+= 'A' - 'a';
		
		*target++		= c;
	}

	return ( count );
}


static unsigned char str_match( char *a, char *b, unsigned char n )
{
	unsigned char	i;

	for ( i = 0; i < n; i++ )
		if ( *a++ != *b++ )
			return ( False );

	return ( True );
}


static void make_full_path( char *target, char *relative, char *current )
{
	unsigned char	i			= 0;
	char 			*t;
	
	t	= target;
	
	//	if the given path from application is absolute path...

	if ( *relative != '/' )
	{
		//	copy relative path
		
		for ( ; i < 128; i++ )
		{		
			if ( !*current )
				break;

			*t++	= *current++;
		}
		
		if ( t != (target + 1) )
			*t++	= '/';

	}
	
	for ( ; i < 128; i++ )
	{
		if ( !*relative )
			break;
	
		*t++	= *relative++;
	}
	
	*t		= 0;

	reform_path( target );
}


static char *reform_path( char *source )
{
	char			target[ MAX_PATH_STRING_LENGTH ];
	char			*s;
	char			*t;
	unsigned short	first_slash;
	unsigned short	last_slash;
	
	s	= source;	
	t	= target;	

	while ( *s )
	{
		s++;
		
		if ( !(first_slash	= first_char_position( s, '/' )) )
			break;
		
		if ( !(*s) )
			break;
		
		if ( (*(s + 0) == '.') && (*(s + 1) != '.') )
			;

		else if ( (*(s + 0) == '.') && (*(s + 1) == '.') )
			t	= target + last_char_position( target, '/' );

		else
		{
			*t++	 = '/';
			t		+= fs_data_copy( (unsigned char *)t, (unsigned char *)s, first_slash );
		}
		
		*t	= 0;
		s	+= first_slash;
	}

	if ( t == target )
		*t++	= '/';
		
	*t	= 0;

	fs_data_copy( (unsigned char *)source, (unsigned char *)target, MAX_PATH_STRING_LENGTH );
	
	return ( source );
}


static void separate_directory_and_file( char *directory_path, char *file_name, char *source )
{
	unsigned short		last_slash;

	fs_data_copy( directory_path, source, MAX_PATH_STRING_LENGTH );
	
	last_slash	= last_char_position( directory_path, '/' );
	
	if ( last_slash == 0 )
	{
		*(directory_path + 0)	= '/';
		*(directory_path + 1)	= 0;
	}
	else
	{
		*(directory_path + last_slash)	= 0;
	}
	
	fs_data_copy( file_name, source + last_slash + 1, 12 );
}


static unsigned short first_char_position( char *path, char mc )
{
	char 			*p;
	char 			c;
	
	p		= path;

	while ( 0 != (c	= *p++) )
		if (c == mc)
			break;

	p--;

	return ( p - path );
}


static unsigned short last_char_position( char *path, char mc )
{
	char 			*p;
	
	p		= path;

	while ( *p++ )
		;

	while ( path < p )
		if (*(--p) == mc)
			break;

	return ( p - path );
}




